@@ -42,11 +42,13 @@ void loadConfig(void)
 	FILE *f;
 
 	// set default config values first
+	config.noDownsampleOnSmpLoad = false;
+	config.disableE8xEffect = false;
 	config.fullScreenStretch = false;
 	config.pattDots = false;
 	config.waveformCenterLine = true;
-	config.a500LowPassFilter = false;
-	config.soundFrequency = 96000;
+	config.filterModel = FILTERMODEL_A1200;
+	config.soundFrequency = 48000;
 	config.rememberPlayMode = false;
 	config.stereoSeparation = 20;
 	config.videoScaleFactor = 2;
@@ -61,12 +63,14 @@ void loadConfig(void)
 	config.autoCloseDiskOp = true;
 	config.vsyncOff = false;
 	config.hwMouse = true;
-	config.sampleLowpass = true;
 	config.startInFullscreen = false;
 	config.pixelFilter = PIXELFILTER_NEAREST;
 	config.integerScaling = true;
 	config.audioInputFrequency = 44100;
-	config.normalizeSampling = true;
+	config.keepEditModeAfterStepPlay = false;
+
+	config.maxSampleLength = 65534;
+	config.reservedSampleOffset = (MOD_SAMPLES+1) * config.maxSampleLength;
 
 #ifndef _WIN32
 	getcwd(oldCwd, PATH_MAX);
@@ -194,6 +198,44 @@ static bool loadProTrackerDotIni(FILE *f)
 			continue;
 		}
 
+		// STEPPLAY_KEEP_EDITMODE
+		else if (!_strnicmp(configLine, "STEPPLAY_KEEP_EDITMODE=", 23))
+		{
+			if (!_strnicmp(&configLine[23], "TRUE", 4))
+				config.keepEditModeAfterStepPlay = true;
+			else if (!_strnicmp(&configLine[23], "FALSE", 5))
+				config.keepEditModeAfterStepPlay = false;
+		}
+
+		// 64K_LIMIT
+		else if (!_strnicmp(configLine, "64K_LIMIT=", 10))
+		{
+			if (!_strnicmp(&configLine[10], "TRUE", 4))
+			{
+				config.maxSampleLength = 65534;
+				config.reservedSampleOffset = (MOD_SAMPLES+1) * config.maxSampleLength;
+			}
+			else if (!_strnicmp(&configLine[10], "FALSE", 5))
+			{
+				config.maxSampleLength = 131070;
+				config.reservedSampleOffset = (MOD_SAMPLES+1) * config.maxSampleLength;
+			}
+		}
+
+		// NO_DWNSMP_ON_SMP_LOAD (no dialog for 2x downsample after >22kHz sample load)
+		else if (!_strnicmp(configLine, "NO_DWNSMP_ON_SMP_LOAD=", 22))
+		{
+			if (!_strnicmp(&configLine[22], "TRUE", 4)) config.noDownsampleOnSmpLoad = true;
+			else if (!_strnicmp(&configLine[22], "FALSE", 5)) config.noDownsampleOnSmpLoad = false;
+		}
+
+		// DISABLE_E8X (Karplus-Strong command)
+		else if (!_strnicmp(configLine, "DISABLE_E8X=", 12))
+		{
+			     if (!_strnicmp(&configLine[12], "TRUE",  4)) config.disableE8xEffect = true;
+			else if (!_strnicmp(&configLine[12], "FALSE", 5)) config.disableE8xEffect = false;
+		}
+
 		// HWMOUSE
 		else if (!_strnicmp(configLine, "HWMOUSE=", 8))
 		{
@@ -201,13 +243,6 @@ static bool loadProTrackerDotIni(FILE *f)
 			else if (!_strnicmp(&configLine[8], "FALSE", 5)) config.hwMouse = false;
 		}
 
-		// SAMPLELOWPASS
-		else if (!_strnicmp(configLine, "SAMPLELOWPASS=", 14))
-		{
-			     if (!_strnicmp(&configLine[14], "TRUE",  4)) config.sampleLowpass = true;
-			else if (!_strnicmp(&configLine[14], "FALSE", 5)) config.sampleLowpass = false;
-		}
-
 		// VSYNCOFF
 		else if (!_strnicmp(configLine, "VSYNCOFF=", 9))
 		{
@@ -375,18 +410,25 @@ static bool loadProTrackerDotIni(FILE *f)
 			}
 		}
 
-		// A500LOWPASSFILTER
-		else if (!_strnicmp(configLine, "A500LOWPASSFILTER=", 18))
+		// FILTERMODEL
+		else if (!_strnicmp(configLine, "FILTERMODEL=", 12))
 		{
-			     if (!_strnicmp(&configLine[18], "TRUE",  4)) config.a500LowPassFilter = true;
-			else if (!_strnicmp(&configLine[18], "FALSE", 5)) config.a500LowPassFilter = false;
+			     if (!_strnicmp(&configLine[12], "A500",  4)) config.filterModel = FILTERMODEL_A500;
+			else if (!_strnicmp(&configLine[12], "A1200", 5)) config.filterModel = FILTERMODEL_A1200;
 		}
 
-		// A4000LOWPASSFILTER (deprecated, same as A500LOWPASSFILTER)
+		// A500LOWPASSFILTER (deprecated, same as A4000LOWPASSFILTER)
+		else if (!_strnicmp(configLine, "A500LOWPASSFILTER=", 18))
+		{
+			     if (!_strnicmp(&configLine[18], "TRUE",  4)) config.filterModel = FILTERMODEL_A500;
+			else if (!_strnicmp(&configLine[18], "FALSE", 5)) config.filterModel = FILTERMODEL_A1200;
+		}
+
+		// A4000LOWPASSFILTER (deprecated)
 		else if (!_strnicmp(configLine, "A4000LOWPASSFILTER=", 19))
 		{
-			     if (!_strnicmp(&configLine[19], "TRUE",  4)) config.a500LowPassFilter = true;
-			else if (!_strnicmp(&configLine[19], "FALSE", 5)) config.a500LowPassFilter = false;
+			     if (!_strnicmp(&configLine[19], "TRUE",  4)) config.filterModel = FILTERMODEL_A500;
+			else if (!_strnicmp(&configLine[19], "FALSE", 5)) config.filterModel = FILTERMODEL_A1200;
 		}
 
 		// SAMPLINGFREQ
@@ -399,13 +441,6 @@ static bool loadProTrackerDotIni(FILE *f)
 			}
 		}
 
-		// NORMALIZESAMPLING
-		else if (!_strnicmp(configLine, "NORMALIZESAMPLING=", 18))
-		{
-			     if (!_strnicmp(&configLine[18], "TRUE",  4)) config.normalizeSampling = true;
-			else if (!_strnicmp(&configLine[18], "FALSE", 5)) config.normalizeSampling = false;
-		}
-
 		// FREQUENCY
 		else if (!_strnicmp(configLine, "FREQUENCY=", 10))
 		{
