@@ -13,13 +13,14 @@ enum
 typedef struct config_t
 {
 	char *defModulesDir, *defSamplesDir;
-	bool waveformCenterLine, pattDots, a500LowPassFilter, compoMode, autoCloseDiskOp, hideDiskOpDates, hwMouse;
+	bool waveformCenterLine, pattDots, compoMode, autoCloseDiskOp, hideDiskOpDates, hwMouse;
 	bool transDel, fullScreenStretch, vsyncOff, modDot, blankZeroFlag, realVuMeters, rememberPlayMode;
-	bool sampleLowpass, startInFullscreen, integerScaling, normalizeSampling;
+	bool startInFullscreen, integerScaling, disableE8xEffect, noDownsampleOnSmpLoad, keepEditModeAfterStepPlay;
 	int8_t stereoSeparation, videoScaleFactor, accidental;
-	uint8_t pixelFilter;
+	uint8_t pixelFilter, filterModel;
 	uint16_t quantizeValue;
-	uint32_t soundFrequency, soundBufferSize, audioInputFrequency;
+	int32_t maxSampleLength;
+	uint32_t soundFrequency, soundBufferSize, audioInputFrequency, reservedSampleOffset;
 } config_t;
 
 extern config_t config; // pt2_config.c
