@@ -27,7 +27,6 @@
 #include "pt2_textout.h"
 #include "pt2_diskop.h"
 #include "pt2_tables.h"
-#include "pt2_palette.h"
 #include "pt2_module_loader.h"
 #include "pt2_audio.h"
 #include "pt2_sampler.h"
@@ -143,7 +142,6 @@ static int8_t findFirst(fileEntry_t *searchRec)
 #else
 	struct dirent *fData;
 	struct stat st;
-	int64_t fSize;
 #endif
 
 	searchRec->nameU = NULL; // this one must be initialized
@@ -173,26 +171,12 @@ static int8_t findFirst(fileEntry_t *searchRec)
 		return LFF_SKIP;
 
 	searchRec->filesize = 0;
-	searchRec->isDir = (fData->d_type == DT_DIR) ? true : false;
+	searchRec->isDir = false;
 
-	if (fData->d_type == DT_UNKNOWN || fData->d_type == DT_LNK)
+	if (stat(fData->d_name, &st) == 0)
 	{
-		if (stat(fData->d_name, &st) == 0)
-		{
-			fSize = (int64_t)st.st_size;
-			searchRec->filesize = (fSize > INT32_MAX) ? -1 : (fSize & 0xFFFFFFFF);
-
-			if ((st.st_mode & S_IFMT) == S_IFDIR)
-				searchRec->isDir = true;
-		}
-	}
-	else if (!searchRec->isDir)
-	{
-		if (stat(fData->d_name, &st) == 0)
-		{
-			fSize = (int64_t)st.st_size;
-			searchRec->filesize = (fSize > INT32_MAX) ? -1 : (fSize & 0xFFFFFFFF);
-		}
+		searchRec->isDir = !!(st.st_mode & S_IFDIR);
+		searchRec->filesize = ((int64_t)st.st_size > INT32_MAX) ? -1 : (int32_t)st.st_size;
 	}
 #endif
 
@@ -227,7 +211,6 @@ static int8_t findNext(fileEntry_t *searchRec)
 #else
 	struct dirent *fData;
 	struct stat st;
-	int64_t fSize;
 #endif
 
 	searchRec->nameU = NULL; // important
@@ -253,26 +236,12 @@ static int8_t findNext(fileEntry_t *searchRec)
 		return LFF_SKIP;
 
 	searchRec->filesize = 0;
-	searchRec->isDir = (fData->d_type == DT_DIR) ? true : false;
+	searchRec->isDir = false;
 
-	if (fData->d_type == DT_UNKNOWN || fData->d_type == DT_LNK)
+	if (stat(fData->d_name, &st) == 0)
 	{
-		if (stat(fData->d_name, &st) == 0)
-		{
-			fSize = (int64_t)st.st_size;
-			searchRec->filesize = (fSize > INT32_MAX) ? -1 : (fSize & 0xFFFFFFFF);
-
-			if ((st.st_mode & S_IFMT) == S_IFDIR)
-				searchRec->isDir = true;
-		}
-	}
-	else if (!searchRec->isDir)
-	{
-		if (stat(fData->d_name, &st) == 0)
-		{
-			fSize = (int64_t)st.st_size;
-			searchRec->filesize = (fSize > INT32_MAX) ? -1 : (fSize & 0xFFFFFFFF);
-		}
+		searchRec->isDir = !!(st.st_mode & S_IFDIR);
+		searchRec->filesize = ((int64_t)st.st_size > INT32_MAX) ? -1 : (int32_t)st.st_size;
 	}
 #endif
 
@@ -545,104 +514,44 @@ void freeDiskOpEntryMem(void)
 	diskop.numEntries = 0;
 }
 
-static bool swapEntries(int32_t a, int32_t b)
+// thanks to aTc for creating this simplified routine for qsort() (I edited it a little bit)
+static int32_t fileEntryCompare(const void *f1, const void *f2)
 {
-	fileEntry_t tmpBuffer;
+	static char fn_1[PATH_MAX+1], fn_2[PATH_MAX+1];
 
-	if (a >= diskop.numEntries || b >= diskop.numEntries)
-		return false;
+	fileEntry_t *fe1 = (fileEntry_t *)f1;
+	fileEntry_t *fe2 = (fileEntry_t *)f2;
 
-	tmpBuffer = diskOpEntry[a];
-	diskOpEntry[a] = diskOpEntry[b];
-	diskOpEntry[b] = tmpBuffer;
+	unicharToAnsi(fn_1, fe1->nameU, PATH_MAX);
+	unicharToAnsi(fn_2, fe2->nameU, PATH_MAX);
 
-	return true;
-}
+	// ".." directories are always sorted first
+	if (fe1->isDir && strcmp(fn_1, "..") == 0) return -1;
+	if (fe2->isDir && strcmp(fn_2, "..") == 0) return  1;
 
-static char *getSortEntry(int32_t entry)
-{
-	char *p;
-	int32_t nameLen;
-	fileEntry_t *dirEntry;
+	/* ProTracker handles names in upper case during sorting,
+	** convert the string case.
+	*/
+	const int32_t fe1_l = (int32_t)strlen(fn_1);
+	const int32_t fe2_l = (int32_t)strlen(fn_2);
+	for (int32_t i = 0; i < fe1_l; i++) fn_1[i] = (char)toupper(fn_1[i]);
+	for (int32_t i = 0; i < fe2_l; i++) fn_2[i] = (char)toupper(fn_2[i]);
 
-	dirEntry = &diskOpEntry[entry];
+	// if both entries are the same type
+	if (fe1->isDir == fe2->isDir)
+		return strcmp(fn_1, fn_2);
 
-	unicharToAnsi(fileNameBuffer, dirEntry->nameU, PATH_MAX);
+	// different types, so one of them is a dir
+	if (fe1->isDir)
+		return -1; // first one is a dir
 
-	nameLen = (int32_t)strlen(fileNameBuffer);
-	if (nameLen == 0)
-		return NULL;
-
-	p = (char *)malloc(nameLen + 2);
-	if (p == NULL)
-		return NULL;
-
-	if (dirEntry->isDir)
-	{
-		if (nameLen == 2 && fileNameBuffer[0] == '.' && fileNameBuffer[1] == '.')
-			p[0] = 0x01; // make ".." directory first priority
-		else
-			p[0] = 0x02; // make second priority
-
-		strcpy(&p[1], fileNameBuffer);
-	}
-	else
-	{
-		strcpy(p, fileNameBuffer);
-	}
-
-	return p;
+	return 1; // second one is a dir
 }
 
 static void sortEntries(void)
 {
-	bool didSwap;
-	char *p1, *p2;
-	uint32_t offset, limit, i;
-
-	if (diskop.numEntries < 2)
-		return; // no need to sort
-
-	offset = diskop.numEntries >> 1;
-	while (offset > 0)
-	{
-		limit = diskop.numEntries - offset;
-		do
-		{
-			didSwap = false;
-			for (i = 0; i < limit; i++)
-			{
-				p1 = getSortEntry(i);
-				p2 = getSortEntry(offset + i);
-
-				if (p1 == NULL || p2 == NULL)
-				{
-					if (p1 != NULL) free(p1);
-					if (p2 != NULL) free(p2);
-					statusOutOfMemory();
-					return;
-				}
-
-				if (_stricmp(p1, p2) > 0)
-				{
-					if (!swapEntries(i , offset + i))
-					{
-						free(p1);
-						free(p2);
-						return;
-					}
-
-					didSwap = true;
-				}
-
-				free(p1);
-				free(p2);
-			}
-		}
-		while (didSwap);
-
-		offset >>= 1;
-	}
+	if (diskop.numEntries >= 2)
+		qsort(diskOpEntry, diskop.numEntries, sizeof (fileEntry_t), fileEntryCompare);
 }
 
 static bool diskOpFillBuffer(void)
