@@ -14,7 +14,7 @@
 #include "pt2_unicode.h"
 #include "pt2_palette.h"
 
-#define PROG_VER_STR "1.18"
+#define PROG_VER_STR "1.51"
 
 #ifdef _WIN32
 #define DIR_DELIMITER '\\'
@@ -36,15 +36,12 @@
 */
 #define VBLANK_HZ 60
 
-/* Scopes are clocked at 64Hz instead of 60Hz to prevent the small +/- Hz 
+/* Scopes are clocked at 64Hz instead of 60Hz to prevent the small +/- Hz
 ** interference from monitors not being exactly 60Hz (and unstable non-vsync mode).
-** Sadly, the scopes might midly flicker from this in some cases.
+** Sadly, the scopes might mildly flicker from this in some cases.
 */
 #define SCOPE_HZ 64
 
-#define AMIGA_PAL_VBLANK_HZ 49.9204092835
-
-#define FONT_BMP_WIDTH 
 #define FONT_CHAR_W 8 // actual data length is 7, includes right spacing (1px column)
 #define FONT_CHAR_H 5
 
@@ -53,11 +50,6 @@
 #define MOD_ORDERS 128
 #define MAX_PATTERNS 100
 
-#define MAX_SAMPLE_LEN 65534
-
-// for NULL pointers
-#define RESERVED_SAMPLE_OFFSET ((31+1) * MAX_SAMPLE_LEN)
-
 #define AMIGA_VOICES 4
 #define SCOPE_WIDTH 40
 #define SCOPE_HEIGHT 33
@@ -67,18 +59,23 @@
 
 #define POSED_LIST_SIZE 12
 
-
-// main crystal oscillator
+// main crystal oscillator for PAL Amiga systems
 #define AMIGA_PAL_XTAL_HZ 28375160
 
-#define PAULA_PAL_CLK (AMIGA_PAL_XTAL_HZ / 8)
-#define CIA_PAL_CLK (AMIGA_PAL_XTAL_HZ / 40)
+#define AMIGA_PAL_CCK_HZ (AMIGA_PAL_XTAL_HZ/8.0)
+
+// nominal framerate in normal PAL videomodes (~49.92Hz)
+#define AMIGA_PAL_VBLANK_HZ (AMIGA_PAL_CCK_HZ / (double)(313*227))
+
+#define PAULA_PAL_CLK AMIGA_PAL_CCK_HZ
+#define CIA_PAL_CLK (AMIGA_PAL_CCK_HZ / 5.0)
 
 #define PAL_PAULA_MIN_SAFE_PERIOD 124
 #define PAL_PAULA_MAX_SAFE_HZ (PAULA_PAL_CLK / (double)PAL_PAULA_MIN_SAFE_PERIOD)
 
 #define FILTERS_BASE_FREQ (PAULA_PAL_CLK / 214.0)
 
+// Amount of video frames. 14 (PT on Amiga) -> 17 (converted from 49.92Hz to 60Hz)
 #define KEYB_REPEAT_DELAY 17
 
 // .MOD types
@@ -103,7 +100,8 @@ enum
 	TEMPFLAG_START = 1,
 	TEMPFLAG_DELAY = 2,
 
-	FILTER_A500 = 1,
+	FILTERMODEL_A1200 = 0,
+	FILTERMODEL_A500 = 1,
 	FILTER_LED_ENABLED = 2,
 
 	NO_CARRY = 0,
@@ -213,7 +211,10 @@ enum
 	TEXT_EDIT_HEX = 2
 };
 
-int8_t *allocMemForAllSamples(void); // pt2_replayer.c
+// pt2_replayer.c
+int8_t *allocMemForAllSamples(void);
+void setReplayerPosToTrackerPos(void);
+// -------------------------
 
 void restartSong(void);
 void resetSong(void);
@@ -224,8 +225,8 @@ void modStop(void);
 void doStopIt(bool resetPlayMode);
 void playPattern(int8_t startRow);
 void modPlay(int16_t patt, int16_t order, int8_t row);
-void modSetSpeed(uint8_t speed);
-void modSetTempo(uint16_t bpm, bool doLockAudio);
+void modSetSpeed(int32_t speed);
+void modSetTempo(int32_t bpm, bool doLockAudio);
 void modFree(void);
 bool setupAudio(void);
 void audioClose(void);
