@@ -16,7 +16,6 @@
 #include "pt2_header.h"
 #include "pt2_helpers.h"
 #include "pt2_visuals.h"
-#include "pt2_palette.h"
 #include "pt2_diskop.h"
 #include "pt2_edit.h"
 #include "pt2_sampler.h"
@@ -29,6 +28,7 @@
 #include "pt2_unicode.h"
 #include "pt2_config.h"
 #include "pt2_sampling.h"
+#include "pt2_chordmaker.h"
 
 #if defined _WIN32 && !defined _DEBUG
 extern bool windowsKeyIsDown;
@@ -43,8 +43,8 @@ void movePatCurLeft(void);
 static bool handleGeneralModes(SDL_Keycode keycode, SDL_Scancode scancode);
 bool handleTextEditMode(SDL_Scancode scancode);
 
-void sampleUpButton(void); // pt_mouse.c
-void sampleDownButton(void); // pt_mouse.c
+void sampleUpButton(void); // pt2_mouse.c
+void sampleDownButton(void); // pt2_mouse.c
 
 void gotoNextMulti(void)
 {
@@ -341,11 +341,21 @@ void keyDownHandler(SDL_Scancode scancode, SDL_Keycode keycode)
 		return;
 	}
 
+	// kludge: allow certain specific key combos to be repeated with the ctrl key
+	const bool nonRepeatAltKeys = keyb.leftAltPressed && scancode != SDL_SCANCODE_DELETE    && scancode != SDL_SCANCODE_RETURN
+	                                                  && scancode != SDL_SCANCODE_BACKSPACE && scancode != SDL_SCANCODE_BACKSLASH
+	                                                  && scancode != SDL_SCANCODE_EQUALS    && scancode != SDL_SCANCODE_MINUS
+	                                                  && scancode <  SDL_SCANCODE_1         && scancode >  SDL_SCANCODE_0;
+
+	// kludge: allow certain specific key combos to be repeated with the alt key
+	const bool nonRepeatCtrlKeys = keyb.leftCtrlPressed && scancode != SDL_SCANCODE_DELETE && scancode != SDL_SCANCODE_RETURN
+	                                                    && scancode != SDL_SCANCODE_BACKSPACE;
+
 	// these keys should not allow to be repeated in keyrepeat mode (caps lock)
-	const bool illegalKeys = keyb.leftAltPressed || keyb.leftAmigaPressed || keyb.leftCtrlPressed
+	const bool nonRepeatKeys = keyb.leftAmigaPressed || nonRepeatAltKeys || nonRepeatCtrlKeys
 		|| scancode == SDL_SCANCODE_LEFT || scancode == SDL_SCANCODE_RIGHT
 		|| scancode == SDL_SCANCODE_UP   || scancode == SDL_SCANCODE_DOWN;
-	if (editor.repeatKeyFlag && keyb.repeatKey && scancode == keyb.lastRepKey && illegalKeys)
+	if (editor.repeatKeyFlag && keyb.repeatKey && scancode == keyb.lastRepKey && nonRepeatKeys)
 		return;
 
 	if (scancode == SDL_SCANCODE_KP_PLUS)
@@ -354,7 +364,7 @@ void keyDownHandler(SDL_Scancode scancode, SDL_Keycode keycode)
 	// TOGGLE FULLSCREEN (should always react)
 	if (scancode == SDL_SCANCODE_F11 && !keyb.leftAltPressed)
 	{
-		toggleFullScreen();
+		toggleFullscreen();
 
 		// prevent fullscreen toggle from firing twice on certain SDL2 Linux ports
 #ifdef __unix__
@@ -699,17 +709,26 @@ void keyDownHandler(SDL_Scancode scancode, SDL_Keycode keycode)
 		{
 			if (keyb.leftCtrlPressed)
 			{
+				const bool audioWasntLocked = !audio.locked;
+				if (audioWasntLocked)
+					lockAudio();
+
 				editor.timingMode ^= 1;
+				updateReplayerTimingMode();
+
 				if (editor.timingMode == TEMPO_MODE_VBLANK)
 				{
 					editor.oldTempo = song->currBPM;
-					modSetTempo(125, true);
+					modSetTempo(125, false);
 				}
 				else
 				{
-					modSetTempo(editor.oldTempo, true);
+					modSetTempo(editor.oldTempo, false);
 				}
 
+				if (audioWasntLocked)
+					unlockAudio();
+
 				ui.updateSongTiming = true;
 			}
 			else if (keyb.shiftPressed)
@@ -718,7 +737,7 @@ void keyDownHandler(SDL_Scancode scancode, SDL_Keycode keycode)
 			}
 			else
 			{
-				toggleA500Filters();
+				toggleFilterModel();
 			}
 		}
 		break;
@@ -807,8 +826,21 @@ void keyDownHandler(SDL_Scancode scancode, SDL_Keycode keycode)
 					editor.stepPlayEnabled = true;
 					editor.stepPlayBackwards = false;
 
-					doStopIt(true);
+					editor.stepPlayLastMode = editor.currMode;
+
+					if (config.keepEditModeAfterStepPlay && editor.stepPlayLastMode == MODE_EDIT)
+						doStopIt(false);
+					else
+						doStopIt(true);
+
 					playPattern(song->currRow);
+
+					if (config.keepEditModeAfterStepPlay && editor.stepPlayLastMode == MODE_EDIT)
+					{
+						pointerSetMode(POINTER_MODE_EDIT, DO_CARRY);
+						editor.playMode = PLAY_MODE_NORMAL;
+						editor.currMode = MODE_EDIT;
+					}
 				}
 			}
 		}
@@ -2257,7 +2289,7 @@ void keyDownHandler(SDL_Scancode scancode, SDL_Keycode keycode)
 #ifdef __APPLE__
 			if (keyb.leftCommandPressed && keyb.leftCtrlPressed)
 			{
-				toggleFullScreen();
+				toggleFullscreen();
 			}
 			else
 #endif
@@ -3198,10 +3230,17 @@ void handleKeyRepeat(SDL_Scancode scancode)
 		{
 			if (ui.editTextFlag)
 			{
-				if (keyb.repeatCounter >= 4)
+				if (keyb.delayCounter >= KEYB_REPEAT_DELAY)
 				{
-					keyb.repeatCounter = 0;
-					textCharPrevious();
+					if (keyb.repeatCounter >= 3)
+					{
+						keyb.repeatCounter = 0;
+						textCharPrevious();
+					}
+				}
+				else
+				{
+					keyb.delayCounter++;
 				}
 			}
 			else
@@ -3248,10 +3287,17 @@ void handleKeyRepeat(SDL_Scancode scancode)
 		{
 			if (ui.editTextFlag)
 			{
-				if (keyb.repeatCounter >= 4)
+				if (keyb.delayCounter >= KEYB_REPEAT_DELAY)
 				{
-					keyb.repeatCounter = 0;
-					textCharNext();
+					if (keyb.repeatCounter >= 3)
+					{
+						keyb.repeatCounter = 0;
+						textCharNext();
+					}
+				}
+				else
+				{
+					keyb.delayCounter++;
 				}
 			}
 			else
@@ -3406,10 +3452,18 @@ void handleKeyRepeat(SDL_Scancode scancode)
 			if (ui.editTextFlag)
 			{
 				// only repeat backspace while editing texts
-				if (keyb.repeatCounter >= 3)
+
+				if (keyb.delayCounter >= KEYB_REPEAT_DELAY)
 				{
-					keyb.repeatCounter = 0;
-					keyDownHandler(scancode, 0);
+					if (keyb.repeatCounter >= 3)
+					{
+						keyb.repeatCounter = 0;
+						keyDownHandler(scancode, 0);
+					}
+				}
+				else
+				{
+					keyb.delayCounter++;
 				}
 			}
 		}
@@ -3430,10 +3484,7 @@ void handleKeyRepeat(SDL_Scancode scancode)
 		break;
 	}
 
-	// repeat keys at 49.92Hz (Amiga PAL) rate
-	const uint64_t keyRepeatDelta = (uint64_t)(((UINT32_MAX + 1.0) * (AMIGA_PAL_VBLANK_HZ / (double)VBLANK_HZ)) + 0.5);
-
-	keyb.repeatFrac += keyRepeatDelta; // 32.32 fixed-point counter
+	keyb.repeatFrac += keyb.repeatDelta; // 32.32 fixed-point counter
 	if (keyb.repeatFrac > 0xFFFFFFFF)
 	{
 		keyb.repeatFrac &= 0xFFFFFFFF;
@@ -3441,11 +3492,27 @@ void handleKeyRepeat(SDL_Scancode scancode)
 	}
 }
 
+static void swapChannel(uint8_t srcCh, uint8_t dstCh)
+{
+	if (srcCh == dstCh)
+		return;
+
+	for (int32_t i = 0; i < MOD_ROWS; i++)
+	{
+		note_t *noteSrc = &song->patterns[song->currPattern][(i * AMIGA_VOICES) + dstCh];
+		note_t noteTmp = song->patterns[song->currPattern][(i * AMIGA_VOICES) + srcCh];
+
+		song->patterns[song->currPattern][(i * AMIGA_VOICES) + srcCh] = *noteSrc;
+		*noteSrc = noteTmp;
+	}
+
+	updateWindowTitle(MOD_IS_MODIFIED);
+	ui.updatePatternData = true;
+}
+
 bool handleGeneralModes(SDL_Keycode keycode, SDL_Scancode scancode)
 {
 	int8_t rawKey;
-	int16_t i;
-	note_t *noteSrc, noteTmp;
 
 	// SAMPLER SCREEN (volume box)
 	if (ui.samplerVolBoxShown && !ui.editTextFlag && scancode == SDL_SCANCODE_ESCAPE)
@@ -3558,22 +3625,22 @@ bool handleGeneralModes(SDL_Keycode keycode, SDL_Scancode scancode)
 			if (ui.changingChordNote == 1)
 			{
 				editor.note1 = rawKey;
-				ui.updateNote1Text = true;
+				ui.updateChordNote1Text = true;
 			}
 			else if (ui.changingChordNote == 2)
 			{
 				editor.note2 = rawKey;
-				ui.updateNote2Text = true;
+				ui.updateChordNote2Text = true;
 			}
 			else if (ui.changingChordNote == 3)
 			{
 				editor.note3 = rawKey;
-				ui.updateNote3Text = true;
+				ui.updateChordNote3Text = true;
 			}
 			else if (ui.changingChordNote == 4)
 			{
 				editor.note4 = rawKey;
-				ui.updateNote4Text = true;
+				ui.updateChordNote4Text = true;
 			}
 
 			ui.changingChordNote = false;
@@ -3719,15 +3786,7 @@ bool handleGeneralModes(SDL_Keycode keycode, SDL_Scancode scancode)
 
 			case SDL_SCANCODE_1:
 			{
-				for (i = 0; i < MOD_ROWS; i++)
-				{
-					noteSrc = &song->patterns[song->currPattern][(i * AMIGA_VOICES) + cursor.channel];
-					noteTmp = song->patterns[song->currPattern][i * AMIGA_VOICES];
-
-					song->patterns[song->currPattern][i * AMIGA_VOICES] = *noteSrc;
-					*noteSrc = noteTmp;
-				}
-
+				swapChannel(0, cursor.channel);
 				editor.swapChannelFlag = false;
 
 				pointerSetPreviousMode();
@@ -3737,15 +3796,7 @@ bool handleGeneralModes(SDL_Keycode keycode, SDL_Scancode scancode)
 
 			case SDL_SCANCODE_2:
 			{
-				for (i = 0; i < MOD_ROWS; i++)
-				{
-					noteSrc = &song->patterns[song->currPattern][(i * AMIGA_VOICES) + cursor.channel];
-					noteTmp = song->patterns[song->currPattern][(i * AMIGA_VOICES) + 1];
-
-					song->patterns[song->currPattern][(i * AMIGA_VOICES) + 1] = *noteSrc;
-					*noteSrc = noteTmp;
-				}
-
+				swapChannel(1, cursor.channel);
 				editor.swapChannelFlag = false;
 
 				pointerSetPreviousMode();
@@ -3755,15 +3806,7 @@ bool handleGeneralModes(SDL_Keycode keycode, SDL_Scancode scancode)
 
 			case SDL_SCANCODE_3:
 			{
-				for (i = 0; i < MOD_ROWS; i++)
-				{
-					noteSrc = &song->patterns[song->currPattern][(i * AMIGA_VOICES) + cursor.channel];
-					noteTmp = song->patterns[song->currPattern][(i * AMIGA_VOICES) + 2];
-
-					song->patterns[song->currPattern][(i * AMIGA_VOICES) + 2] = *noteSrc;
-					*noteSrc = noteTmp;
-				}
-
+				swapChannel(2, cursor.channel);
 				editor.swapChannelFlag = false;
 
 				pointerSetPreviousMode();
@@ -3773,15 +3816,7 @@ bool handleGeneralModes(SDL_Keycode keycode, SDL_Scancode scancode)
 
 			case SDL_SCANCODE_4:
 			{
-				for (i = 0; i < MOD_ROWS; i++)
-				{
-					noteSrc = &song->patterns[song->currPattern][(i * AMIGA_VOICES) + cursor.channel];
-					noteTmp = song->patterns[song->currPattern][(i * AMIGA_VOICES) + 3];
-
-					song->patterns[song->currPattern][(i * AMIGA_VOICES) + 3] = *noteSrc;
-					*noteSrc = noteTmp;
-				}
-
+				swapChannel(3, cursor.channel);
 				editor.swapChannelFlag = false;
 
 				pointerSetPreviousMode();
@@ -3937,12 +3972,6 @@ bool handleGeneralModes(SDL_Keycode keycode, SDL_Scancode scancode)
 
 				pointerSetPreviousMode();
 				setPrevStatusMessage();
-
-				editor.errorMsgActive = true;
-				editor.errorMsgBlock = true;
-				editor.errorMsgCounter = 0;
-
-				setErrPointer();
 			}
 			break;
 
@@ -3988,7 +4017,7 @@ void handleTextEditInputChar(char textChar)
 
 				textMarkerMoveRight();
 			}
-			else if ((textChar >= '0' && textChar <= '9') || (textChar >= 'A' && textChar <= 'F'))
+			else if ((textChar >= '0' && textChar <= '9') || (textChar >= 'a' && textChar <= 'f'))
 			{
 				if (ui.dstPos == 14) // hack for sample mix text
 				{
@@ -4072,15 +4101,28 @@ void handleTextEditInputChar(char textChar)
 				else if (textChar <= 'f')
 					textChar -= 'a'-10;
 
-				if (ui.numBits == 16)
+				if (ui.numBits == 17)
 				{
-					*ui.numPtr16 &= ~(0xF000 >> (ui.dstPos << 2));
-					*ui.numPtr16 |= (textChar << (12 - (ui.dstPos << 2)));
+					*ui.numPtr32 &= ~(0xF0000 >> (ui.dstPos << 2));
+					*ui.numPtr32 |= textChar << (16 - (ui.dstPos << 2));
+				}
+				else if (ui.numBits == 16)
+				{
+					if (ui.force32BitNumPtr)
+					{
+						*ui.numPtr32 &= ~(0xF000 >> (ui.dstPos << 2));
+						*ui.numPtr32 |= textChar << (12 - (ui.dstPos << 2));
+					}
+					else
+					{
+						*ui.numPtr16 &= ~(0xF000 >> (ui.dstPos << 2));
+						*ui.numPtr16 |= textChar << (12 - (ui.dstPos << 2));
+					}
 				}
 				else if (ui.numBits == 8)
 				{
 					*ui.numPtr8 &= ~(0xF0 >> (ui.dstPos << 2));
-					*ui.numPtr8 |= (textChar << (4 - (ui.dstPos << 2)));
+					*ui.numPtr8 |= textChar << (4 - (ui.dstPos << 2));
 				}
 
 				textMarkerMoveRight();
@@ -4202,7 +4244,7 @@ bool handleTextEditMode(SDL_Scancode scancode)
 					keyb.delayCounter = 0;
 
 				keyb.repeatKey = true;
-				keyb.delayKey = false;
+				keyb.delayKey = true;
 
 				updateTextObject(ui.editObject);
 			}
@@ -4315,8 +4357,21 @@ bool handleTextEditMode(SDL_Scancode scancode)
 					editor.stepPlayEnabled = true;
 					editor.stepPlayBackwards = true;
 
-					doStopIt(true);
+					editor.stepPlayLastMode = editor.currMode;
+
+					if (config.keepEditModeAfterStepPlay && editor.stepPlayLastMode == MODE_EDIT)
+						doStopIt(false);
+					else
+						doStopIt(true);
+
 					playPattern((song->currRow - 1) & 0x3F);
+
+					if (config.keepEditModeAfterStepPlay && editor.stepPlayLastMode == MODE_EDIT)
+					{
+						pointerSetMode(POINTER_MODE_EDIT, DO_CARRY);
+						editor.playMode = PLAY_MODE_NORMAL;
+						editor.currMode = MODE_EDIT;
+					}
 				}
 			}
 		}
