@@ -19,7 +19,6 @@
 #include <sys/stat.h>
 #include "pt2_header.h"
 #include "pt2_helpers.h"
-#include "pt2_palette.h"
 #include "pt2_keyboard.h"
 #include "pt2_textout.h"
 #include "pt2_mouse.h"
@@ -37,6 +36,7 @@
 #include "pt2_bmp.h"
 #include "pt2_sync.h"
 #include "pt2_sampling.h"
+#include "pt2_hpc.h"
 
 #define CRASH_TEXT "Oh no!\nThe ProTracker 2 clone has crashed...\n\nA backup .mod was hopefully " \
                    "saved to the current module directory.\n\nPlease report this bug if you can.\n" \
@@ -44,7 +44,7 @@
 
 module_t *song = NULL; // globalized
 
-static bool backupMadeAfterCrash;
+static bool backupMadeAfterCrash, didDropFile;
 
 #ifdef _WIN32
 #define SYSMSG_FILE_ARG (WM_USER + 1)
@@ -58,9 +58,10 @@ HHOOK g_hKeyboardHook;
 static HWND hWnd_to;
 static HANDLE oneInstHandle, hMapFile;
 static LPCTSTR sharedMemBuf;
-static TCHAR sharedHwndName[] = TEXT("Local\\PTCloneHwnd");
-static TCHAR sharedFileName[] = TEXT("Local\\PTCloneFilename");
+static TCHAR sharedHwndName[] = TEXT("Local\\PT2CloneHwnd");
+static TCHAR sharedFileName[] = TEXT("Local\\PT2CloneFilename");
 static bool handleSingleInstancing(int32_t argc, char **argv);
+static void closeSingleInstancing(void);
 static void handleSysMsg(SDL_Event inputEvent);
 #endif
 
@@ -163,6 +164,11 @@ int main(int argc, char *argv[])
 #endif
 
 #ifdef _WIN32
+
+#ifndef _MSC_VER
+	SetProcessDPIAware();
+#endif
+
 	if (!SDL_HasSSE())
 	{
 		showErrorMsgBox("Your computer's processor doesn't have the SSE+SSE2 instruction set\n" \
@@ -177,7 +183,6 @@ int main(int argc, char *argv[])
 		return 0;
 	}
 
-	setupWin32Usleep();
 	disableWasapi(); // disable problematic WASAPI SDL2 audio driver on Windows (causes clicks/pops sometimes...)
 	                 // 13.03.2020: This is still needed with SDL 2.0.12...
 #endif
@@ -198,6 +203,7 @@ int main(int argc, char *argv[])
 	}
 
 	SDL_SetThreadPriority(SDL_THREAD_PRIORITY_HIGH);
+	SDL_EnableScreenSaver(); // allow screensaver to activate
 
 	/* Text input is started by default in SDL2, turn it off to remove ~2ms spikes per key press.
 	** We manuallay start it again when someone clicks on a text edit box, and stop it when done.
@@ -256,9 +262,10 @@ int main(int argc, char *argv[])
 	SDL_EventState(SDL_SYSWMEVENT, SDL_ENABLE);
 #endif
 
-	setupPerfFreq();
+	hpc_Init();
+	hpc_SetDurationInHz(&video.vblankHpc, VBLANK_HZ);
 
-	if (!setupAudio() || !unpackBMPs())
+	if (!initKaiserTable() || !setupAudio() || !unpackBMPs())
 	{
 		cleanUp();
 		SDL_Quit();
@@ -317,14 +324,14 @@ int main(int argc, char *argv[])
 	SDL_ShowWindow(video.window);
 
 	if (config.startInFullscreen)
-		toggleFullScreen();
+		toggleFullscreen();
 
 	changePathToHome(); // set path to home/user-dir now
 	diskOpSetInitPath(); // set path to custom path in config (if present)
 
 	SDL_EventState(SDL_DROPFILE, SDL_ENABLE);
 
-	setupWaitVBL();
+	hpc_ResetEndTime(&video.vblankHpc);
 	while (editor.programRunning)
 	{
 		sinkVisualizerBars();
@@ -365,20 +372,11 @@ static void handleInput(void)
 			else if (event.window.event == SDL_WINDOWEVENT_SHOWN)
 				video.windowHidden = false;
 
-			if (video.vsync60HzPresent)
-			{
-				/* if we minimize the window and vsync is present, vsync is temporarily turned off.
-				** recalc waitVBL() vars so that it can sleep properly in said mode.
-				*/
-				if (event.window.event == SDL_WINDOWEVENT_MINIMIZED ||
-					event.window.event == SDL_WINDOWEVENT_FOCUS_LOST)
-				{
-					setupWaitVBL();
-				}
-			}
+			// reset vblank end time if we minimize window
+			if (event.window.event == SDL_WINDOWEVENT_MINIMIZED || event.window.event == SDL_WINDOWEVENT_FOCUS_LOST)
+				hpc_ResetEndTime(&video.vblankHpc);
 		}
 
-
 #ifdef _WIN32
 		handleSysMsg(event);
 #endif
@@ -402,7 +400,10 @@ static void handleInput(void)
 		{
 			loadDroppedFile(event.drop.file, (uint32_t)strlen(event.drop.file), false, true);
 			SDL_free(event.drop.file);
-			SDL_RaiseWindow(video.window); // set window focus
+
+			// kludge: allow focus-clickthrough after drag-n-drop
+			SDL_SetHint(SDL_HINT_MOUSE_FOCUS_CLICKTHROUGH, "1");
+			didDropFile = true;
 		}
 		if (event.type == SDL_QUIT)
 		{
@@ -423,20 +424,23 @@ static void handleInput(void)
 
 			if (!ui.askScreenShown && ui.introScreenShown)
 			{
-				if (!ui.clearScreenShown && !ui.diskOpScreenShown)
+				if (!ui.clearScreenShown && !ui.diskOpScreenShown && !editor.errorMsgActive)
 					statusAllRight();
 
 				ui.introScreenShown = false;
 			}
+
+			// kludge: we drag-n-dropped a file before this mouse click release, restore focus-clickthrough mode
+			if (didDropFile)
+			{
+				didDropFile = false;
+				SDL_SetHint(SDL_HINT_MOUSE_FOCUS_CLICKTHROUGH, "0");
+			}
 		}
 		else if (event.type == SDL_MOUSEBUTTONDOWN)
 		{
-			if (ui.sampleMarkingPos == -1 &&
-				!ui.forceSampleDrag && !ui.forceVolDrag &&
-				!ui.forceSampleEdit)
-			{
+			if (ui.sampleMarkingPos == -1 && !ui.forceSampleDrag && !ui.forceVolDrag && !ui.forceSampleEdit)
 				mouseButtonDownHandler(event.button.button);
-			}
 		}
 
 		if (ui.throwExit)
@@ -467,6 +471,10 @@ static bool initializeVars(void)
 
 	editor.repeatKeyFlag = (SDL_GetModState() & KMOD_CAPS) ? true : false;
 
+	// set key repeat rate to 49.9204Hz (Amiga PAL vblank rate)
+	const double dVblankHzRatio = AMIGA_PAL_VBLANK_HZ / (double)VBLANK_HZ;
+	keyb.repeatDelta = (uint64_t)floor((UINT32_MAX+1.0) * dVblankHzRatio);
+
 	strcpy(editor.mixText, "MIX 01+02 TO 03");
 
 	// allocate some memory
@@ -476,13 +484,10 @@ static bool initializeVars(void)
 
 	config.defModulesDir = (char *)calloc(PATH_MAX + 1, sizeof (char));
 	config.defSamplesDir = (char *)calloc(PATH_MAX + 1, sizeof (char));
-	editor.tempSample = (int8_t *)calloc(MAX_SAMPLE_LEN, 1);
+	editor.tempSample = (int8_t *)calloc(131070, 1);
 
-	if (config.defModulesDir == NULL || config.defSamplesDir == NULL ||
-		editor.tempSample == NULL)
-	{
+	if (config.defModulesDir == NULL || config.defSamplesDir == NULL || editor.tempSample == NULL)
 		goto oom;
-	}
 
 	turnOffVoices();
 
@@ -504,7 +509,7 @@ static bool initializeVars(void)
 	editor.oldNote4 = 36;
 	editor.tuningVol = 32;
 	editor.sampleVol = 100;
-	editor.tuningNote = 24;
+	editor.tuningNote = 24; // C-3
 	editor.metroSpeed = 4;
 	editor.editMoveAdd = 1;
 	editor.initialTempo = 125;
@@ -643,39 +648,37 @@ static bool checkIfAppWasTranslocated(int argc, char **argv)
 #ifdef _WIN32
 static void disableWasapi(void)
 {
-	const char *audioDriver;
-	int32_t i, numAudioDrivers;
-
 	// disable problematic WASAPI SDL2 audio driver on Windows (causes clicks/pops sometimes...)
 
-	numAudioDrivers = SDL_GetNumAudioDrivers();
-	for (i = 0; i < numAudioDrivers; i++)
+	const int32_t numAudioDrivers = SDL_GetNumAudioDrivers();
+	if (numAudioDrivers <= 1)
+		return;
+
+	// look for directsound and enable it if found
+	for (int32_t i = 0; i < numAudioDrivers; i++)
 	{
-		audioDriver = SDL_GetAudioDriver(i);
+		const char *audioDriver = SDL_GetAudioDriver(i);
 		if (audioDriver != NULL && strcmp("directsound", audioDriver) == 0)
 		{
 			SDL_setenv("SDL_AUDIODRIVER", "directsound", true);
 			audio.rescanAudioDevicesSupported = false;
-			break;
+			return;
 		}
 	}
 
-	if (i == numAudioDrivers)
+	// directsound is not available, try winmm
+	for (int32_t i = 0; i < numAudioDrivers; i++)
 	{
-		// directsound is not available, try winmm
-		for (i = 0; i < numAudioDrivers; i++)
+		const char *audioDriver = SDL_GetAudioDriver(i);
+		if (audioDriver != NULL && strcmp("winmm", audioDriver) == 0)
 		{
-			audioDriver = SDL_GetAudioDriver(i);
-			if (audioDriver != NULL && strcmp("winmm", audioDriver) == 0)
-			{
-				SDL_setenv("SDL_AUDIODRIVER", "winmm", true);
-				audio.rescanAudioDevicesSupported = false;
-				break;
-			}
+			SDL_setenv("SDL_AUDIODRIVER", "winmm", true);
+			audio.rescanAudioDevicesSupported = false;
+			return;
 		}
 	}
 
-	// maybe we didn't find directsound or winmm, let's use wasapi after all then...
+	// we didn't find directsound or winmm, let's use wasapi after all...
 }
 
 static void makeSureDirIsProgramDir(void)
@@ -776,6 +779,7 @@ static bool handleSingleInstancing(int32_t argc, char **argv)
 		if (sharedMemBuf != NULL)
 		{
 			memcpy(&hWnd_to, sharedMemBuf, sizeof (HWND));
+
 			UnmapViewOfFile(sharedMemBuf);
 			sharedMemBuf = NULL;
 			CloseHandle(hMapFile);
@@ -788,11 +792,12 @@ static bool handleSingleInstancing(int32_t argc, char **argv)
 				if (sharedMemBuf != NULL)
 				{
 					strcpy((char *)sharedMemBuf, argv[1]);
+
 					UnmapViewOfFile(sharedMemBuf);
 					sharedMemBuf = NULL;
 
 					SendMessage(hWnd_to, SYSMSG_FILE_ARG, 0, 0);
-					SDL_Delay(80); // wait a bit to make sure first instance received msg
+					Sleep(80); // wait a bit to make sure first instance received msg
 
 					CloseHandle(hMapFile);
 					hMapFile = NULL;
@@ -813,31 +818,39 @@ static bool handleSingleInstancing(int32_t argc, char **argv)
 
 static void handleSysMsg(SDL_Event inputEvent)
 {
-	SDL_SysWMmsg *wmMsg;
+	if (inputEvent.type != SDL_SYSWMEVENT)
+		return;
 
-	if (inputEvent.type == SDL_SYSWMEVENT)
+	SDL_SysWMmsg *wmMsg = inputEvent.syswm.msg;
+	if (wmMsg->subsystem == SDL_SYSWM_WINDOWS && wmMsg->msg.win.msg == SYSMSG_FILE_ARG)
 	{
-		wmMsg = inputEvent.syswm.msg;
-		if (wmMsg->subsystem == SDL_SYSWM_WINDOWS && wmMsg->msg.win.msg == SYSMSG_FILE_ARG)
+		hMapFile = OpenFileMapping(FILE_MAP_READ, FALSE, sharedFileName);
+		if (hMapFile != NULL)
 		{
-			hMapFile = OpenFileMapping(FILE_MAP_READ, FALSE, sharedFileName);
-			if (hMapFile != NULL)
+			sharedMemBuf = (LPTSTR)MapViewOfFile(hMapFile, FILE_MAP_READ, 0, 0, ARGV_SHARED_MEM_MAX_LEN);
+			if (sharedMemBuf != NULL)
 			{
-				sharedMemBuf = (LPTSTR)MapViewOfFile(hMapFile, FILE_MAP_READ, 0, 0, ARGV_SHARED_MEM_MAX_LEN);
-				if (sharedMemBuf != NULL)
-				{
-					loadDroppedFile((char *)sharedMemBuf, (uint32_t)strlen(sharedMemBuf), true, true);
-					UnmapViewOfFile(sharedMemBuf);
-					sharedMemBuf = NULL;
-				}
+				loadDroppedFile((char *)sharedMemBuf, (uint32_t)strlen(sharedMemBuf), true, true);
 
-				CloseHandle(hMapFile);
-				hMapFile = NULL;
+				UnmapViewOfFile(sharedMemBuf);
+				sharedMemBuf = NULL;
 			}
+
+			CloseHandle(hMapFile);
+			hMapFile = NULL;
 		}
 	}
 }
 
+void closeSingleInstancing(void)
+{
+	if (oneInstHandle != NULL)
+	{
+		CloseHandle(oneInstHandle);
+		oneInstHandle = NULL;
+	}
+}
+
 static LONG WINAPI exceptionHandler(EXCEPTION_POINTERS *ptr)
 #else
 static void exceptionHandler(int32_t signal)
@@ -892,16 +905,20 @@ static void cleanUp(void) // never call this inside the main loop!
 	videoClose();
 	freeSprites();
 	freeAudioDeviceList(); // pt2_sampling.c
+	freeKaiserTable(); // pt2_sampling.c
 
 	if (config.defModulesDir != NULL) free(config.defModulesDir);
 	if (config.defSamplesDir != NULL) free(config.defSamplesDir);
 	if (editor.tempSample != NULL) free(editor.tempSample);
 
 #ifdef _WIN32
-	freeWin32Usleep();
 #ifndef _DEBUG
 	UnhookWindowsHookEx(g_hKeyboardHook);
 #endif
-	if (oneInstHandle != NULL) CloseHandle(oneInstHandle);
+	if (oneInstHandle != NULL)
+	{
+		CloseHandle(oneInstHandle);
+		oneInstHandle = NULL;
+	}
 #endif
 }
