@@ -1,4 +1,4 @@
-// C port of ProTracker 2.3D's replayer by 8bitbubsy, slightly modified.
+// C port of ProTracker 2.3D's replayer (with some modifications, but still accurate)
 
 // for finding memory leaks in debug mode with Visual Studio 
 #if defined _DEBUG && defined _MSC_VER
@@ -13,7 +13,6 @@
 #include "pt2_header.h"
 #include "pt2_audio.h"
 #include "pt2_helpers.h"
-#include "pt2_palette.h"
 #include "pt2_tables.h"
 #include "pt2_module_loader.h"
 #include "pt2_config.h"
@@ -23,11 +22,12 @@
 #include "pt2_scopes.h"
 #include "pt2_sync.h"
 
-static bool posJumpAssert, pBreakFlag, updateUIPositions, modHasBeenPlayed;
+static bool posJumpAssert, pBreakFlag, modRenderDone;
+static bool doStopSong; // from F00 (Set Speed)
 static int8_t pBreakPosition, oldRow, modPattern;
-static uint8_t pattDelTime, setBPMFlag, lowMask = 0xFF, pattDelTime2, oldSpeed;
+static uint8_t pattDelTime, lowMask = 0xFF, pattDelTime2;
 static int16_t modOrder, oldPattern, oldOrder;
-static uint16_t modBPM, oldBPM;
+static int32_t modBPM, oldBPM, oldSpeed, ciaSetBPM;
 
 static const uint8_t funkTable[16] = // EFx (FunkRepeat/InvertLoop)
 {
@@ -35,29 +35,28 @@ static const uint8_t funkTable[16] = // EFx (FunkRepeat/InvertLoop)
 	0x10, 0x13, 0x16, 0x1A, 0x20, 0x2B, 0x40, 0x80
 };
 
-int8_t *allocMemForAllSamples(void)
+void setReplayerPosToTrackerPos(void)
 {
-	/* Allocate memoru for all sample data blocks.
-	**
-	** We need three extra sample slots:
-	** The 1st is extra safety padding since setting a Paula length of 0
-	** results in reading (1+65535)*2 bytes. The 2nd and 3rd (64K*2 = 1x 128K)
-	** are reserved for NULL pointers. This is needed for emulating a PT quirk.
-	**
-	** We have a padding of 4 bytes at the end for length=0 quirk safety.
-	**
-	** PS: I don't really know if it's possible for ProTracker to set a Paula
-	** length of 0, but I fully support this Paula behavior just in case.
-	*/
-	const size_t allocLen = ((MOD_SAMPLES + 3) * MAX_SAMPLE_LEN) + 4;
+	if (song == NULL)
+		return;
 
-	return (int8_t *)calloc(1, allocLen);
+	modPattern = (int8_t)song->currPattern;
+	modOrder = song->currOrder;
+	song->row = song->currRow;
+	song->tick = 0;
 }
 
-void modSetSpeed(uint8_t speed)
+int8_t *allocMemForAllSamples(void)
 {
-	song->speed = speed;
-	song->currSpeed = speed;
+	// allocate memory for all sample data blocks (+ 2 extra, for quirk + safety)
+	const size_t allocLen = (MOD_SAMPLES + 2) * config.maxSampleLength;
+
+	return (int8_t *)calloc(allocLen, 1);
+}
+
+void modSetSpeed(int32_t speed)
+{
+	song->currSpeed = song->speed = speed;
 	song->tick = 0;
 }
 
@@ -81,16 +80,17 @@ void doStopIt(bool resetPlayMode)
 
 	if (song != NULL)
 	{
-		for (int32_t i = 0; i < AMIGA_VOICES; i++)
+		moduleChannel_t *c = song->channels;
+		for (int32_t i = 0; i < AMIGA_VOICES; i++, c++)
 		{
-			moduleChannel_t *c = &song->channels[i];
-
 			c->n_wavecontrol = 0;
 			c->n_glissfunk = 0;
 			c->n_finetune = 0;
 			c->n_loopcount = 0;
 		}
 	}
+
+	doStopSong = false; // just in case this flag was stuck from command F00 (stop song)
 }
 
 void setPattern(int16_t pattern)
@@ -146,7 +146,7 @@ static void updateFunk(moduleChannel_t *ch)
 	{
 		ch->n_funkoffset = 0;
 
-		if (ch->n_loopstart != NULL && ch->n_wavestart != NULL) // non-PT2 bug fix
+		if (ch->n_loopstart != NULL && ch->n_wavestart != NULL) // ProTracker bugfix
 		{
 			if (++ch->n_wavestart >= ch->n_loopstart + (ch->n_replen << 1))
 				ch->n_wavestart = ch->n_loopstart;
@@ -206,16 +206,34 @@ static void setTremoloControl(moduleChannel_t *ch)
 	ch->n_wavecontrol = ((ch->n_cmd & 0xF) << 4) | (ch->n_wavecontrol & 0xF);
 }
 
+/* This is a little used effect, despite being present in original ProTracker.
+** E8x was sometimes entirely replaced with code used for demo fx syncing in
+** demo mod players, so it can be turned off by looking at DISABLE_E8X in
+** protracker.ini if you so desire.
+*/
 static void karplusStrong(moduleChannel_t *ch)
 {
-	/* This effect is definitely the least used PT effect there is!
-	** It trashes (filters) the sample data.
-	** The reason I'm not implementing it is because a lot of songs used
-	** E8x for syncing to demos/intros, and because I have never ever
-	** seen this effect being used intentionally.
-	*/
+	int8_t a, b;
 
-	(void)ch;
+	if (config.disableE8xEffect)
+		return;
+
+	if (ch->n_loopstart == NULL)
+		return; // ProTracker bugfix
+
+	int8_t *ptr8 = ch->n_loopstart;
+	int16_t end = ((ch->n_replen * 2) & 0xFFFF) - 2;
+	do
+	{
+		a = ptr8[0];
+		b = ptr8[1];
+		*ptr8++ = (a + b) >> 1;
+	}
+	while (--end >= 0);
+
+	a = ptr8[0];
+	b = ch->n_loopstart[0];
+	*ptr8 = (a + b) >> 1;
 }
 
 static void doRetrg(moduleChannel_t *ch)
@@ -250,17 +268,16 @@ static void retrigNote(moduleChannel_t *ch)
 
 static void volumeSlide(moduleChannel_t *ch)
 {
-	uint8_t cmd = ch->n_cmd & 0xFF;
-
-	if ((cmd & 0xF0) == 0)
+	uint8_t param = ch->n_cmd & 0xFF;
+	if ((param & 0xF0) == 0)
 	{
-		ch->n_volume -= cmd & 0x0F;
+		ch->n_volume -= param & 0x0F;
 		if (ch->n_volume < 0)
 			ch->n_volume = 0;
 	}
 	else
 	{
-		ch->n_volume += cmd >> 4;
+		ch->n_volume += param >> 4;
 		if (ch->n_volume > 64)
 			ch->n_volume = 64;
 	}
@@ -309,6 +326,7 @@ static void funkIt(moduleChannel_t *ch)
 	if (song->tick == 0)
 	{
 		ch->n_glissfunk = ((ch->n_cmd & 0xF) << 4) | (ch->n_glissfunk & 0xF);
+
 		if ((ch->n_glissfunk & 0xF0) > 0)
 			updateFunk(ch);
 	}
@@ -316,7 +334,10 @@ static void funkIt(moduleChannel_t *ch)
 
 static void positionJump(moduleChannel_t *ch)
 {
-	modOrder = (ch->n_cmd & 0xFF) - 1; // 0xFF (B00) jumps to pat 0
+	// original PT doesn't do this check, but we have to
+	if (editor.playMode != PLAY_MODE_PATTERN || (editor.currMode == MODE_RECORD && editor.recordMode != RECORD_PATT))
+		modOrder = (ch->n_cmd & 0xFF) - 1; // B00 results in -1, but it safely jumps to order 0
+
 	pBreakPosition = 0;
 	posJumpAssert = true;
 }
@@ -341,20 +362,15 @@ static void setSpeed(moduleChannel_t *ch)
 {
 	if ((ch->n_cmd & 0xFF) > 0)
 	{
-		song->tick = 0;
-
 		if (editor.timingMode == TEMPO_MODE_VBLANK || (ch->n_cmd & 0xFF) < 32)
 			modSetSpeed(ch->n_cmd & 0xFF);
 		else
-			setBPMFlag = ch->n_cmd & 0xFF; // CIA doesn't refresh its registers until the next interrupt, so change it later
+			ciaSetBPM = ch->n_cmd & 0xFF; // the CIA chip doesn't use its new timer value until the next interrupt, so change it later
 	}
 	else
 	{
-		editor.songPlaying = false;
-		editor.playMode = PLAY_MODE_NORMAL;
-		editor.currMode = MODE_IDLE;
-
-		pointerResetThreadSafe(); // set gray mouse cursor
+		// F00 - stop song
+		doStopSong = true;
 	}
 }
 
@@ -363,9 +379,7 @@ static void arpeggio(moduleChannel_t *ch)
 	uint8_t arpTick, arpNote;
 	const int16_t *periods;
 
-	assert(song->tick < 32);
-	arpTick = arpTickTable[song->tick]; // 0, 1, 2
-
+	arpTick = song->tick % 3; // 0, 1, 2
 	if (arpTick == 1)
 	{
 		arpNote = (uint8_t)(ch->n_cmd >> 4);
@@ -401,7 +415,7 @@ static void portaUp(moduleChannel_t *ch)
 	ch->n_period -= (ch->n_cmd & 0xFF) & lowMask;
 	lowMask = 0xFF;
 
-	if ((ch->n_period & 0xFFF) < 113)
+	if ((ch->n_period & 0xFFF) < 113) // PT BUG: sign removed before comparison, underflow not clamped!
 		ch->n_period = (ch->n_period & 0xF000) | 113;
 
 	paulaSetPeriod(ch->n_chanindex, ch->n_period & 0xFFF);
@@ -420,7 +434,11 @@ static void portaDown(moduleChannel_t *ch)
 
 static void filterOnOff(moduleChannel_t *ch)
 {
-	setLEDFilter(!(ch->n_cmd & 1), false);
+	if (song->tick == 0) // added this (just pointless to call this during all ticks!)
+	{
+		const bool filterOn = (ch->n_cmd & 1) ^ 1;
+		setLEDFilter(filterOn, false);
+	}
 }
 
 static void finePortaUp(moduleChannel_t *ch)
@@ -545,23 +563,20 @@ static void vibrato2(moduleChannel_t *ch)
 	const uint8_t vibratoPos = (ch->n_vibratopos >> 2) & 0x1F;
 	const uint8_t vibratoType = ch->n_wavecontrol & 3;
 
-	if (vibratoType == 0)
+	if (vibratoType == 0) // sine
 	{
 		vibratoData = vibratoTable[vibratoPos];
 	}
-	else
+	else if (vibratoType == 1) // ramp
 	{
-		if (vibratoType == 1)
-		{
-			if (ch->n_vibratopos < 128)
-				vibratoData = vibratoPos << 3;
-			else
-				vibratoData = 255 - (vibratoPos << 3);
-		}
+		if (ch->n_vibratopos < 128)
+			vibratoData = vibratoPos << 3;
 		else
-		{
-			vibratoData = 255;
-		}
+			vibratoData = 255 - (vibratoPos << 3);
+	}
+	else // square
+	{
+		vibratoData = 255;
 	}
 
 	vibratoData = (vibratoData * (ch->n_vibratocmd & 0xF)) >> 7;
@@ -612,23 +627,20 @@ static void tremolo(moduleChannel_t *ch)
 	const uint8_t tremoloPos = (ch->n_tremolopos >> 2) & 0x1F;
 	const uint8_t tremoloType = (ch->n_wavecontrol >> 4) & 3;
 
-	if (tremoloType == 0)
+	if (tremoloType == 0) // sine
 	{
 		tremoloData = vibratoTable[tremoloPos];
 	}
-	else
+	else if (tremoloType == 1) // ramp
 	{
-		if (tremoloType == 1)
-		{
-			if (ch->n_vibratopos < 128) // PT bug, should've been ch->n_tremolopos
-				tremoloData = tremoloPos << 3;
-			else
-				tremoloData = 255 - (tremoloPos << 3);
-		}
+		if (ch->n_vibratopos < 128) // PT bug, should've been ch->n_tremolopos
+			tremoloData = tremoloPos << 3;
 		else
-		{
-			tremoloData = 255;
-		}
+			tremoloData = 255 - (tremoloPos << 3);
+	}
+	else // square
+	{
+		tremoloData = 255;
 	}
 
 	tremoloData = ((uint16_t)tremoloData * (ch->n_tremolocmd & 0xF)) >> 6;
@@ -658,7 +670,8 @@ static void sampleOffset(moduleChannel_t *ch)
 
 	uint16_t newOffset = ch->n_sampleoffset << 7;
 
-	if ((int16_t)newOffset < ch->n_length)
+	// this signed test is the reason for the 9xx "sample >64kB = silence" bug
+	if ((int16_t)newOffset < (int16_t)ch->n_length)
 	{
 		ch->n_length -= newOffset;
 		ch->n_start += newOffset << 1;
@@ -671,62 +684,89 @@ static void sampleOffset(moduleChannel_t *ch)
 
 static void E_Commands(moduleChannel_t *ch)
 {
-	const uint8_t cmd = (ch->n_cmd & 0xF0) >> 4;
-	switch (cmd)
+	const uint8_t ecmd = (ch->n_cmd & 0x00F0) >> 4;
+	switch (ecmd)
 	{
-		case 0x0: filterOnOff(ch);       break;
-		case 0x1: finePortaUp(ch);       break;
-		case 0x2: finePortaDown(ch);     break;
-		case 0x3: setGlissControl(ch);   break;
-		case 0x4: setVibratoControl(ch); break;
-		case 0x5: setFineTune(ch);       break;
-		case 0x6: jumpLoop(ch);          break;
-		case 0x7: setTremoloControl(ch); break;
-		case 0x8: karplusStrong(ch);     break;
+		case 0x0: filterOnOff(ch);       return;
+		case 0x1: finePortaUp(ch);       return;
+		case 0x2: finePortaDown(ch);     return;
+		case 0x3: setGlissControl(ch);   return;
+		case 0x4: setVibratoControl(ch); return;
+		case 0x5: setFineTune(ch);       return;
+		case 0x6: jumpLoop(ch);          return;
+		case 0x7: setTremoloControl(ch); return;
+		case 0x8: karplusStrong(ch);     return;
+		case 0xE: patternDelay(ch);      return;
 		default: break;
 	}
 
 	if (editor.muted[ch->n_chanindex])
 		return;
 
-	switch (cmd)
+	switch (ecmd)
 	{
-		case 0x9: retrigNote(ch);     break;
-		case 0xA: volumeFineUp(ch);   break;
-		case 0xB: volumeFineDown(ch); break;
-		case 0xC: noteCut(ch);        break;
-		case 0xD: noteDelay(ch);      break;
-		case 0xE: patternDelay(ch);   break;
-		case 0xF: funkIt(ch);         break;
+		case 0x9: retrigNote(ch);     return;
+		case 0xA: volumeFineUp(ch);   return;
+		case 0xB: volumeFineDown(ch); return;
+		case 0xC: noteCut(ch);        return;
+		case 0xD: noteDelay(ch);      return;
+		case 0xF: funkIt(ch);         return;
 		default: break;
 	}
 }
 
 static void checkMoreEffects(moduleChannel_t *ch)
 {
-	switch ((ch->n_cmd & 0xF00) >> 8)
+	const uint8_t cmd = (ch->n_cmd & 0x0F00) >> 8;
+	switch (cmd)
 	{
-		case 0x9: sampleOffset(ch); break;
-		case 0xB: positionJump(ch); break;
-
-		case 0xC:
-		{
-			if (!editor.muted[ch->n_chanindex])
-				volumeChange(ch);
-		}
-		break;
-
-		case 0xD: patternBreak(ch); break;
-		case 0xE: E_Commands(ch);   break;
-		case 0xF: setSpeed(ch);     break;
-
-		default:
-		{
-			if (!editor.muted[ch->n_chanindex])
-				paulaSetPeriod(ch->n_chanindex, ch->n_period);
-		}
-		break;
+		case 0x9: sampleOffset(ch); return; // note the returns here, not breaks!
+		case 0xB: positionJump(ch); return;
+		case 0xD: patternBreak(ch); return;
+		case 0xE: E_Commands(ch);   return;
+		case 0xF: setSpeed(ch);     return;
+		default: break;
 	}
+
+	if (editor.muted[ch->n_chanindex])
+		return;
+
+	if (cmd == 0xC)
+	{
+		volumeChange(ch);
+		return;
+	}
+
+	paulaSetPeriod(ch->n_chanindex, ch->n_period);
+}
+
+static void chkefx2(moduleChannel_t *ch)
+{
+	updateFunk(ch);
+
+	if ((ch->n_cmd & 0xFFF) == 0)
+		return;
+
+	const uint8_t cmd = (ch->n_cmd & 0x0F00) >> 8;
+	switch (cmd)
+	{
+		case 0x0: arpeggio(ch);            return; // note the returns here, not breaks!
+		case 0x1: portaUp(ch);             return;
+		case 0x2: portaDown(ch);           return;
+		case 0x3: tonePortamento(ch);      return;
+		case 0x4: vibrato(ch);             return;
+		case 0x5: tonePlusVolSlide(ch);    return;
+		case 0x6: vibratoPlusVolSlide(ch); return;
+		case 0xE: E_Commands(ch);          return;
+		default: break;
+	}
+
+	paulaSetPeriod(ch->n_chanindex, ch->n_period);
+
+	if (cmd == 0x7)
+		tremolo(ch);
+	else if (cmd == 0xA)
+		volumeSlide(ch);
 }
 
 static void checkEffects(moduleChannel_t *ch)
@@ -734,41 +774,17 @@ static void checkEffects(moduleChannel_t *ch)
 	if (editor.muted[ch->n_chanindex])
 		return;
 
-	updateFunk(ch);
+	chkefx2(ch);
 
-	const uint8_t effect = (ch->n_cmd & 0xF00) >> 8;
-	if ((ch->n_cmd & 0xFFF) > 0)
-	{
-		switch (effect)
-		{
-			case 0x0: arpeggio(ch);            break;
-			case 0x1: portaUp(ch);             break;
-			case 0x2: portaDown(ch);           break;
-			case 0x3: tonePortamento(ch);      break;
-			case 0x4: vibrato(ch);             break;
-			case 0x5: tonePlusVolSlide(ch);    break;
-			case 0x6: vibratoPlusVolSlide(ch); break;
-			case 0xE: E_Commands(ch);          break;
-
-			case 0x7:
-			{
-				paulaSetPeriod(ch->n_chanindex, ch->n_period);
-				tremolo(ch);
-			}
-			break;
-
-			case 0xA:
-			{
-				paulaSetPeriod(ch->n_chanindex, ch->n_period);
-				volumeSlide(ch);
-			}
-			break;
-
-			default: paulaSetPeriod(ch->n_chanindex, ch->n_period); break;
-		}
-	}
-
-	if (effect != 0x7)
+	/* This is not very clear in the original PT replayer code,
+	** but the tremolo effect skips chkefx2()'s return address
+	** in the stack so that it jumps to checkEffects()'s return
+	** address instead of ending up here. In other words, volume
+	** is not updated here after tremolo (it's done inside the
+	** tremolo routine itself).
+	*/
+	const uint8_t cmd = (ch->n_cmd & 0x0F00) >> 8;
+	if (cmd != 0x7)
 		paulaSetVolume(ch->n_chanindex, ch->n_volume);
 }
 
@@ -858,10 +874,10 @@ static void playVoice(moduleChannel_t *ch)
 		ch->n_start = &song->sampleData[s->offset];
 		ch->n_finetune = s->fineTune & 0xF;
 		ch->n_volume = s->volume;
-		ch->n_length = s->length >> 1;
-		ch->n_replen = s->loopLength >> 1;
+		ch->n_length = (uint16_t)(s->length >> 1);
+		ch->n_replen = (uint16_t)(s->loopLength >> 1);
 
-		const uint16_t repeat = s->loopStart >> 1;
+		const uint16_t repeat = (uint16_t)(s->loopStart >> 1);
 		if (repeat > 0)
 		{
 			ch->n_loopstart = ch->n_start + (repeat << 1);
@@ -874,9 +890,9 @@ static void playVoice(moduleChannel_t *ch)
 			ch->n_wavestart = ch->n_start;
 		}
 
-		// non-PT2 quirk
+		// non-PT2 requirement (set safe sample space for uninitialized voices - f.ex. "the ultimate beeper.mod")
 		if (ch->n_length == 0)
-			ch->n_loopstart = ch->n_wavestart = &song->sampleData[RESERVED_SAMPLE_OFFSET]; // 128K reserved sample
+			ch->n_loopstart = ch->n_wavestart = &song->sampleData[config.reservedSampleOffset]; // 128K reserved sample
 	}
 
 	if ((ch->n_note & 0xFFF) > 0)
@@ -888,7 +904,7 @@ static void playVoice(moduleChannel_t *ch)
 		}
 		else
 		{
-			cmd = (ch->n_cmd & 0xF00) >> 8;
+			cmd = (ch->n_cmd & 0x0F00) >> 8;
 			if (cmd == 3 || cmd == 5)
 			{
 				setVUMeterHeight(ch);
@@ -912,8 +928,36 @@ static void playVoice(moduleChannel_t *ch)
 	}
 }
 
+static void updateUIPositions(void)
+{
+	// don't update UI under MOD2WAV/PAT2SMP rendering
+	if (editor.isWAVRendering || editor.isSMPRendering)
+		return;
+
+	song->currRow = song->row;
+	song->currOrder = modOrder;
+	song->currPattern = modPattern;
+
+	uint16_t *currPatPtr = &song->header.order[modOrder];
+	editor.currPatternDisp = currPatPtr;
+	editor.currPosEdPattDisp = currPatPtr;
+	editor.currPatternDisp = currPatPtr;
+	editor.currPosEdPattDisp = currPatPtr;
+
+	ui.updateSongPos = true;
+	ui.updateSongPattern = true;
+	ui.updateCurrPattText = true;
+	ui.updatePatternData = true;
+
+	if (ui.posEdScreenShown)
+		ui.updatePosEd = true;
+}
+
 static void nextPosition(void)
 {
+	if (editor.isSMPRendering)
+		modRenderDone = true;
+
 	song->row = pBreakPosition;
 	pBreakPosition = 0;
 	posJumpAssert = false;
@@ -923,13 +967,33 @@ static void nextPosition(void)
 	{
 		if (editor.stepPlayEnabled)
 		{
-			doStopIt(true);
+			if (config.keepEditModeAfterStepPlay && editor.stepPlayLastMode == MODE_EDIT)
+			{
+				doStopIt(false);
+
+				pointerSetMode(POINTER_MODE_EDIT, DO_CARRY);
+				editor.playMode = PLAY_MODE_NORMAL;
+				editor.currMode = MODE_EDIT;
+			}
+			else
+			{
+				doStopIt(true);
+			}
 
 			editor.stepPlayEnabled = false;
 			editor.stepPlayBackwards = false;
 
-			if (!editor.isWAVRendering && !editor.isSMPRendering)
+			if (editor.stepPlayLastMode == MODE_EDIT || editor.stepPlayLastMode == MODE_IDLE)
+			{
+				song->row &= 0x3F;
 				song->currRow = song->row;
+			}
+			else
+			{
+				// if we were playing, set replayer row to tracker row (stay in sync)
+				song->currRow &= 0x3F;
+				song->row = song->currRow;
+			}
 
 			return;
 		}
@@ -938,98 +1002,100 @@ static void nextPosition(void)
 		if (modOrder >= song->header.numOrders)
 		{
 			modOrder = 0;
-			modHasBeenPlayed = true;
 
 			if (config.compoMode) // stop song for music competitions playing
 			{
 				doStopIt(true);
 				turnOffVoices();
 
-				song->currOrder = 0;
-				song->currRow = song->row = 0;
-				song->currPattern = modPattern = (int8_t)song->header.order[0];
+				modOrder = 0;
+				modPattern = (int8_t)song->header.order[modOrder];
+				song->row = 0;
 
-				editor.currPatternDisp = &song->currPattern;
-				editor.currPosEdPattDisp = &song->currPattern;
-				editor.currPatternDisp = &song->currPattern;
-				editor.currPosEdPattDisp = &song->currPattern;
-
-				if (ui.posEdScreenShown)
-					ui.updatePosEd = true;
-
-				ui.updateSongPos = true;
-				ui.updateSongPattern = true;
-				ui.updateCurrPattText = true;
+				updateUIPositions();
 			}
+
+			if (editor.isWAVRendering)
+				modRenderDone = true;
 		}
 
 		modPattern = (int8_t)song->header.order[modOrder];
 		if (modPattern > MAX_PATTERNS-1)
 			modPattern = MAX_PATTERNS-1;
-
-		updateUIPositions = true;
 	}
 }
 
-bool intMusic(void)
+static void increasePlaybackTimer(void)
 {
-	uint8_t i;
-	uint16_t *patt;
-	moduleChannel_t *c;
+	// the timer is not counting in "play pattern" mode
+	if (editor.playMode != PLAY_MODE_PATTERN && modBPM >= 32 && modBPM <= 255)
+		editor.musicTime64 += musicTimeTab64[modBPM-32];
+}
 
-	if (modBPM > 0)
-		editor.musicTime += (65536 / modBPM); // for playback counter
+static void setCurrRowToVisited(void) // for MOD2WAV
+{
+	if (editor.isWAVRendering)
+		editor.rowVisitTable[(modOrder * MOD_ROWS) + song->row] = true;
+}
 
-	if (updateUIPositions)
+static bool renderEndCheck(void) // for MOD2WAV/PAT2SMP
+{
+	if (!editor.isWAVRendering && !editor.isSMPRendering)
+		return true; // we're not doing MOD2WAV/PAT2SMP
+
+	bool noPatternDelay = pattDelTime2 == 0;
+	if (noPatternDelay && song->tick == song->speed-1)
 	{
-		updateUIPositions = false;
-
-		if (!editor.isWAVRendering && !editor.isSMPRendering)
+		if (editor.isSMPRendering)
 		{
-			if (editor.playMode != PLAY_MODE_PATTERN)
-			{
-				song->currOrder = modOrder;
-				song->currPattern = modPattern;
-
-				patt = &song->header.order[modOrder];
-				editor.currPatternDisp = patt;
-				editor.currPosEdPattDisp = patt;
-				editor.currPatternDisp = patt;
-				editor.currPosEdPattDisp = patt;
-
-				if (ui.posEdScreenShown)
-					ui.updatePosEd = true;
-
-				ui.updateSongPos = true;
-				ui.updateSongPattern = true;
-				ui.updateCurrPattText = true;
-			}
+			if (modRenderDone)
+				return false; // we're done rendering
+		}
+		
+		if (editor.isWAVRendering)
+		{
+			bool rowVisited = editor.rowVisitTable[(modOrder * MOD_ROWS) + song->row];
+			if (rowVisited || modRenderDone)
+				return false; // we're done rendering
 		}
 	}
 
-	// PT quirk: CIA refreshes its timer values on the next interrupt, so do the real tempo change here
-	if (setBPMFlag != 0)
+	return true;
+}
+
+bool intMusic(void) // replayer ticker
+{
+	// quirk: CIA BPM changes are delayed by one tick in PT, so handle previous tick's BPM change now
+	if (ciaSetBPM != -1)
 	{
-		modSetTempo(setBPMFlag, false);
-		setBPMFlag = 0;
+		const int32_t newBPM = ciaSetBPM;
+		modSetTempo(newBPM, false);
+		ciaSetBPM = -1;
 	}
 
-	if (editor.isWAVRendering && song->tick == 0)
-		editor.rowVisitTable[(modOrder * MOD_ROWS) + song->row] = true;
+	increasePlaybackTimer();
 
 	if (!editor.stepPlayEnabled)
 		song->tick++;
 
-	if (song->tick >= song->speed || editor.stepPlayEnabled)
+	bool readNewNote = false;
+	if ((uint32_t)song->tick >= (uint32_t)song->speed)
 	{
 		song->tick = 0;
+		readNewNote = true;
+	}
 
-		if (pattDelTime2 == 0)
+	if (readNewNote || editor.stepPlayEnabled) // tick 0
+	{
+		if (pattDelTime2 == 0) // no pattern delay, time to read note data
 		{
-			for (i = 0; i < AMIGA_VOICES; i++)
-			{
-				c = &song->channels[i];
+			setCurrRowToVisited(); // for MOD2WAV/PAT2SMP
+			updateUIPositions(); // update current song positions in UI
 
+			// read note data and trigger voices
+			moduleChannel_t *c = song->channels;
+			for (int32_t i = 0; i < AMIGA_VOICES; i++, c++)
+			{
 				playVoice(c);
 				paulaSetVolume(i, c->n_volume);
 
@@ -1038,22 +1104,18 @@ bool intMusic(void)
 				paulaSetLength(i, c->n_replen);
 			}
 		}
-		else
+		else // pattern delay is on-going
 		{
-			for (i = 0; i < AMIGA_VOICES; i++)
-				checkEffects(&song->channels[i]);
-		}
-
-		if (!editor.isWAVRendering && !editor.isSMPRendering)
-		{
-			song->currRow = song->row;
-			ui.updatePatternData = true;
+			moduleChannel_t *c = song->channels;
+			for (int32_t i = 0; i < AMIGA_VOICES; i++, c++)
+				checkEffects(c);
 		}
 
+		// increase row
 		if (!editor.stepPlayBackwards)
 		{
 			song->row++;
-			song->rowsCounter++;
+			song->rowsCounter++; // for MOD2WAV's progress bar
 		}
 
 		if (pattDelTime > 0)
@@ -1062,10 +1124,15 @@ bool intMusic(void)
 			pattDelTime = 0;
 		}
 
+		// undo row increase if pattern delay is on-going
 		if (pattDelTime2 > 0)
 		{
-			if (--pattDelTime2 > 0)
+			pattDelTime2--;
+			if (pattDelTime2 > 0)
+			{
 				song->row--;
+				song->rowsCounter--; // for MOD2WAV's progress bar
+			}
 		}
 
 		if (pBreakFlag)
@@ -1075,50 +1142,71 @@ bool intMusic(void)
 			pBreakFlag = false;
 		}
 
-		if (editor.blockMarkFlag)
-			ui.updateStatusText = true;
-
+		// step-play handling
 		if (editor.stepPlayEnabled)
 		{
-			doStopIt(true);
+			if (config.keepEditModeAfterStepPlay && editor.stepPlayLastMode == MODE_EDIT)
+			{
+				doStopIt(false);
 
-			song->currRow = song->row & 0x3F;
-			ui.updatePatternData = true;
+				pointerSetMode(POINTER_MODE_EDIT, DO_CARRY);
+				editor.playMode = PLAY_MODE_NORMAL;
+				editor.currMode = MODE_EDIT;
+			}
+			else
+			{
+				doStopIt(true);
+			}
+
+			if (editor.stepPlayLastMode == MODE_EDIT || editor.stepPlayLastMode == MODE_IDLE)
+			{
+				song->row &= 0x3F;
+				song->currRow = song->row;
+			}
+			else
+			{
+				// if we were playing, set replayer row to tracker row (stay in sync)
+				song->currRow &= 0x3F;
+				song->row = song->currRow;
+			}
 
 			editor.stepPlayEnabled = false;
 			editor.stepPlayBackwards = false;
-			ui.updatePatternData = true;
 
+			ui.updatePatternData = true;
 			return true;
 		}
 
 		if (song->row >= MOD_ROWS || posJumpAssert)
-		{
-			if (editor.isSMPRendering)
-				modHasBeenPlayed = true;
-
 			nextPosition();
-		}
 
-		if (editor.isWAVRendering && !pattDelTime2 && editor.rowVisitTable[(modOrder * MOD_ROWS) + song->row])
-			modHasBeenPlayed = true;
+		// for pattern block mark feature
+		if (editor.blockMarkFlag)
+			ui.updateStatusText = true;
 	}
-	else
+	else // tick > 0 (handle effects)
 	{
-		for (i = 0; i < AMIGA_VOICES; i++)
-			checkEffects(&song->channels[i]);
+		moduleChannel_t *c = song->channels;
+		for (int32_t i = 0; i < AMIGA_VOICES; i++, c++)
+			checkEffects(c);
 
 		if (posJumpAssert)
 			nextPosition();
 	}
 
-	if ((editor.isSMPRendering || editor.isWAVRendering) && modHasBeenPlayed && song->tick == song->speed-1)
+	// command F00 = stop song, do it here (so that the scopes are updated properly)
+	if (doStopSong)
 	{
-		modHasBeenPlayed = false;
-		return false;
+		doStopSong = false;
+
+		editor.songPlaying = false;
+		editor.playMode = PLAY_MODE_NORMAL;
+		editor.currMode = MODE_IDLE;
+
+		pointerResetThreadSafe(); // set normal gray mouse pointer
 	}
 
-	return true;
+	return renderEndCheck(); // MOD2WAV/PAT2SMP listens to the return value (true = not done yet)
 }
 
 void modSetPattern(uint8_t pattern)
@@ -1179,18 +1267,15 @@ void modSetPos(int16_t order, int16_t row)
 		ui.updateStatusText = true;
 }
 
-void modSetTempo(uint16_t bpm, bool doLockAudio)
+void modSetTempo(int32_t bpm, bool doLockAudio)
 {
-	uint32_t smpsPerTick;
-
-	if (bpm < 32)
+	if (bpm < 32 || bpm > 255)
 		return;
 
 	const bool audioWasntLocked = !audio.locked;
-
 	if (doLockAudio && audioWasntLocked)
 		lockAudio();
-
+	
 	modBPM = bpm;
 	if (!editor.isSMPRendering && !editor.isWAVRendering)
 	{
@@ -1200,19 +1285,18 @@ void modSetTempo(uint16_t bpm, bool doLockAudio)
 
 	bpm -= 32; // 32..255 -> 0..223
 
+	int64_t samplesPerTick64;
 	if (editor.isSMPRendering)
-		smpsPerTick = editor.pat2SmpHQ ? audio.bpmTab28kHz[bpm] : audio.bpmTab22kHz[bpm];
-	else if (editor.isWAVRendering)
-		smpsPerTick = audio.bpmTabMod2Wav[bpm];
+		samplesPerTick64 = editor.pat2SmpHQ ? audio.bpmTable28kHz[bpm] : audio.bpmTable20kHz[bpm];
 	else
-		smpsPerTick = audio.bpmTab[bpm];
+		samplesPerTick64 = audio.bpmTable[bpm];
 
-	mixerSetSamplesPerTick(smpsPerTick);
+	audio.samplesPerTick64 = samplesPerTick64;
 
 	// calculate tick time length for audio/video sync timestamp
-	const uint64_t tickTimeLen64 = audio.tickTimeLengthTab[bpm];
+	const uint64_t tickTimeLen64 = audio.tickLengthTable[bpm];
 	const uint32_t tickTimeLen = tickTimeLen64 >> 32;
-	const uint32_t tickTimeLenFrac = tickTimeLen64 & 0xFFFFFFFF;
+	const uint32_t tickTimeLenFrac = (uint32_t)tickTimeLen64;
 
 	setSyncTickTimeLen(tickTimeLen, tickTimeLenFrac);
 
@@ -1227,10 +1311,9 @@ void modStop(void)
 
 	if (song != NULL)
 	{
-		for (int32_t i = 0; i < AMIGA_VOICES; i++)
+		moduleChannel_t *c = song->channels;
+		for (int32_t i = 0; i < AMIGA_VOICES; i++, c++)
 		{
-			moduleChannel_t *c = &song->channels[i];
-
 			c->n_wavecontrol = 0;
 			c->n_glissfunk = 0;
 			c->n_finetune = 0;
@@ -1243,7 +1326,16 @@ void modStop(void)
 	pattDelTime2 = 0;
 	pBreakPosition = 0;
 	posJumpAssert = false;
-	modHasBeenPlayed = true;
+	modRenderDone = true;
+
+	/* The replayer is one tick ahead (unfortunately), so if the user was to stop the mod at the previous tick
+	** before a position jump (pattern loop, pattern break, position jump, row 63->0 transition, etc),
+	** it would be possible for the replayer to be at another order/pattern than the tracker.
+	** Let's set the replayer state to the tracker state on mod stop, to fix possible confusion.
+	*/
+	setReplayerPosToTrackerPos();
+
+	doStopSong = false; // just in case this flag was stuck from command F00 (stop song)
 }
 
 void playPattern(int8_t startRow)
@@ -1251,10 +1343,15 @@ void playPattern(int8_t startRow)
 	if (!editor.stepPlayEnabled)
 		pointerSetMode(POINTER_MODE_PLAY, DO_CARRY);
 
-	mixerClearSampleCounter();
-
+	audio.tickSampleCounter64 = 0; // zero tick sample counter so that it will instantly initiate a tick
 	song->currRow = song->row = startRow & 0x3F;
-	song->tick = song->speed;
+
+	if (!editor.stepPlayEnabled)
+		song->tick = song->speed-1;
+	else
+		song->tick = 0;
+
+	ciaSetBPM = -1; // fix possibly stuck "set BPM" flag
 
 	editor.playMode = PLAY_MODE_PATTERN;
 	editor.currMode = MODE_PLAY;
@@ -1290,9 +1387,12 @@ void modPlay(int16_t patt, int16_t order, int8_t row)
 {
 	uint8_t oldPlayMode, oldMode;
 
+	const bool audioWasntLocked = !audio.locked;
+	if (audioWasntLocked)
+		lockAudio();
+
 	doStopIt(false);
 	turnOffVoices();
-	mixerClearSampleCounter();
 
 	if (row != -1)
 	{
@@ -1343,11 +1443,20 @@ void modPlay(int16_t patt, int16_t order, int8_t row)
 	editor.playMode = oldPlayMode;
 	editor.currMode = oldMode;
 
-	song->tick = song->speed;
-	modHasBeenPlayed = false;
+	song->tick = song->speed-1;
+	ciaSetBPM = -1; // fix possibly stuck "set BPM" flag
+
+	modRenderDone = false;
 	editor.songPlaying = true;
 	editor.didQuantize = false;
-	editor.musicTime = 0;
+
+	if (editor.playMode != PLAY_MODE_PATTERN)
+		editor.musicTime64 = 0; // don't reset playback counter in "play/rec pattern" mode
+
+	audio.tickSampleCounter64 = 0; // zero tick sample counter so that it will instantly initiate a tick
+
+	if (audioWasntLocked)
+		unlockAudio();
 
 	if (!editor.isSMPRendering && !editor.isWAVRendering)
 	{
@@ -1381,7 +1490,7 @@ void clearSong(void)
 	editor.f9Pos = 48;
 	editor.f10Pos = 63;
 
-	editor.musicTime = 0;
+	editor.musicTime64 = 0;
 
 	editor.metroFlag = false;
 	editor.currSample = 0;
@@ -1443,7 +1552,7 @@ void clearSamples(void)
 		memset(s->text, 0, sizeof (s->text));
 	}
 
-	memset(song->sampleData, 0, (MOD_SAMPLES + 1) * MAX_SAMPLE_LEN);
+	memset(song->sampleData, 0, (MOD_SAMPLES + 1) * config.maxSampleLength);
 
 	editor.currSample = 0;
 	editor.hiLowInstr = 0;
@@ -1462,6 +1571,8 @@ void clearAll(void)
 {
 	clearSamples();
 	clearSong();
+
+	updateWindowTitle(MOD_NOT_MODIFIED);
 }
 
 void modFree(void)
@@ -1500,7 +1611,7 @@ void restartSong(void) // for the beginning of MOD2WAV/PAT2SMP
 
 	editor.playMode = PLAY_MODE_NORMAL;
 	editor.blockMarkFlag = false;
-	audio.forceMixerOff = true;
+	audio.forceSoundCardSilence = true;
 
 	song->row = 0;
 	song->currRow = 0;
@@ -1534,9 +1645,9 @@ void resetSong(void) // only call this after storeTempVariables() has been calle
 
 	turnOffVoices();
 
-	memset((int8_t *)editor.vuMeterVolumes,0, sizeof (editor.vuMeterVolumes));
+	memset((int8_t *)editor.vuMeterVolumes,     0, sizeof (editor.vuMeterVolumes));
 	memset((int8_t *)editor.realVuMeterVolumes, 0, sizeof (editor.realVuMeterVolumes));
-	memset((int8_t *)editor.spectrumVolumes, 0, sizeof (editor.spectrumVolumes));
+	memset((int8_t *)editor.spectrumVolumes,    0, sizeof (editor.spectrumVolumes));
 
 	memset(song->channels, 0, sizeof (song->channels));
 	for (uint8_t i = 0; i < AMIGA_VOICES; i++)
@@ -1562,6 +1673,6 @@ void resetSong(void) // only call this after storeTempVariables() has been calle
 	doStopIt(true);
 
 	song->tick = 0;
-	modHasBeenPlayed = false;
-	audio.forceMixerOff = false;
+	modRenderDone = false;
+	audio.forceSoundCardSilence = false;
 }
