@@ -8,6 +8,7 @@
 #include <stdint.h>
 #include <stdbool.h>
 #include "pt2_header.h"
+#include "pt2_hpc.h"
 
 // for .WAV sample loading/saving
 typedef struct wavHeader_t
@@ -57,12 +58,11 @@ typedef struct moduleHeader_t
 typedef struct moduleSample_t
 {
 	volatile int8_t *volumeDisp;
-	volatile uint16_t *lengthDisp, *loopStartDisp, *loopLengthDisp;
+	volatile int32_t *lengthDisp, *loopStartDisp, *loopLengthDisp;
 	char text[22 + 1];
 	int8_t volume;
 	uint8_t fineTune;
-	uint16_t length, loopStart, loopLength;
-	int32_t offset;
+	int32_t offset, length, loopStart, loopLength;
 } moduleSample_t;
 
 typedef struct moduleChannel_t
@@ -87,7 +87,7 @@ typedef struct module_t
 	bool loaded, modified;
 	int8_t *sampleData;
 
-	volatile uint8_t tick, speed;
+	volatile int32_t tick, speed;
 
 	int8_t row; // used for different things, so must not be internal to replayer
 
@@ -98,8 +98,8 @@ typedef struct module_t
 
 	// for pattern viewer
 	int8_t currRow;
-	uint8_t currSpeed;
-	uint16_t currOrder, currPattern, currBPM;
+	int32_t currSpeed, currBPM;
+	uint16_t currOrder, currPattern;
 
 	// for MOD2WAV progress bar
 	uint32_t rowsCounter, rowsInTotal;
@@ -111,7 +111,7 @@ typedef struct keyb_t
 	bool shiftPressed, leftCtrlPressed, leftAltPressed;
 	bool leftCommandPressed, leftAmigaPressed, keypadEnterPressed;
 	uint8_t repeatCounter, delayCounter;
-	uint64_t repeatFrac;
+	uint64_t repeatDelta, repeatFrac;
 	SDL_Scancode lastRepKey, lastKey;
 } keyb_t;
 
@@ -120,7 +120,7 @@ typedef struct mouse_t
 	volatile bool setPosFlag, resetCursorColorFlag;
 	bool buttonWaiting, leftButtonPressed, rightButtonPressed;
 	uint8_t repeatCounter, buttonWaitCounter;
-	int32_t x, y, lastMouseX, setPosX, setPosY, lastGUIButton, prevX, prevY;
+	int32_t rawX, rawY, x, y, lastMouseX, setPosX, setPosY, lastGUIButton, prevX, prevY;
 	int32_t lastSmpFilterButton, lastSamplingButton;
 	uint32_t buttonState;
 } mouse_t;
@@ -131,6 +131,7 @@ typedef struct video_t
 	int32_t renderX, renderY, renderW, renderH, displayW, displayH;
 	int32_t xScale, yScale;
 	float fMouseXMul, fMouseYMul;
+	hpc_t vblankHpc;
 	SDL_PixelFormat *pixelFormat;
 	uint32_t *frameBuffer, *frameBufferUnaligned;
 
@@ -153,7 +154,7 @@ typedef struct editor_t
 	volatile uint16_t *quantizeValueDisp, *metroSpeedDisp, *metroChannelDisp, *sampleVolDisp;
 	volatile uint16_t *vol1Disp, *vol2Disp, *currEditPatternDisp, *currPosDisp, *currPatternDisp;
 	volatile uint16_t *currPosEdPattDisp, *currLengthDisp, *lpCutOffDisp, *hpCutOffDisp;
-	volatile uint16_t *samplePosDisp, *chordLengthDisp;
+	volatile int32_t *samplePosDisp, *chordLengthDisp;
 
 	char mixText[16];
 	char *entryNameTmp, *currPath, *dropTempFileName;
@@ -168,18 +169,18 @@ typedef struct editor_t
 
 	int8_t smpRedoFinetunes[MOD_SAMPLES], smpRedoVolumes[MOD_SAMPLES], multiModeNext[4], trackPattFlag;
 	int8_t *smpRedoBuffer[MOD_SAMPLES], *tempSample, currSample, recordMode, sampleFrom, sampleTo, autoInsSlot;
-	int8_t hiLowInstr, note1, note2, note3, note4, oldNote1, oldNote2, oldNote3, oldNote4;
+	int8_t hiLowInstr, note1, note2, note3, note4, oldNote1, oldNote2, oldNote3, oldNote4, stepPlayLastMode;
 	uint8_t playMode, currMode, tuningChan, tuningVol, errorMsgCounter, buffFromPos, buffToPos;
 	uint8_t blockFromPos, blockToPos, timingMode, f6Pos, f7Pos, f8Pos, f9Pos, f10Pos, keyOctave, pNoteFlag;
 	uint8_t tuningNote, resampleNote, initialTempo, initialSpeed, editMoveAdd;
 
-	int16_t *pat2SmpBuf, modulateSpeed;
-	uint16_t metroSpeed, metroChannel, sampleVol, samplePos, chordLength;
-	uint16_t effectMacros[10], oldTempo, currPlayNote, vol1, vol2, lpCutOff, hpCutOff;
-	uint16_t smpRedoLoopStarts[MOD_SAMPLES], smpRedoLoopLengths[MOD_SAMPLES], smpRedoLengths[MOD_SAMPLES];
-	int32_t modulatePos, modulateOffset, markStartOfs, markEndOfs, pat2SmpPos;
-	uint32_t musicTime, vblankTimeLen, vblankTimeLenFrac;
-	double dPerfFreq, dPerfFreqMulMicro;
+	int16_t modulateSpeed;
+	uint16_t metroSpeed, metroChannel, sampleVol;
+	uint16_t effectMacros[10], currPlayNote, vol1, vol2, lpCutOff, hpCutOff;
+	int32_t smpRedoLoopStarts[MOD_SAMPLES], smpRedoLoopLengths[MOD_SAMPLES], smpRedoLengths[MOD_SAMPLES];
+	int32_t oldTempo, modulatePos, modulateOffset, markStartOfs, markEndOfs, pat2SmpPos, samplePos, chordLength;
+	uint64_t musicTime64;
+	double *dPat2SmpBuf;
 	note_t trackBuffer[MOD_ROWS], cmdsBuffer[MOD_ROWS], blockBuffer[MOD_ROWS];
 	note_t patternBuffer[MOD_ROWS * AMIGA_VOICES], undoBuffer[MOD_ROWS * AMIGA_VOICES];
 	SDL_Thread *mod2WavThread, *pat2SmpThread;
@@ -213,6 +214,7 @@ typedef struct ui_t
 
 	bool changingSamplingNote;
 
+	bool force32BitNumPtr;
 	int8_t *numPtr8, tmpDisp8, pointerMode, editOpScreen, editTextType, askScreenType;
 	int8_t visualizerMode, previousPointerMode, forceVolDrag, changingChordNote;
 	uint8_t numLen, numBits;
@@ -229,10 +231,10 @@ typedef struct ui_t
 	bool updateMixText, updatePosText, updateModText, updateVolText;
 
 	// edit op. #4 (sample chord editor)
-	bool updateLengthText, updateNote1Text, updateNote2Text;
-	bool updateNote3Text, updateNote4Text;
+	bool updateChordLengthText, updateChordNote1Text, updateChordNote2Text;
+	bool updateChordNote3Text, updateChordNote4Text;
 
-	//sampler
+	// sampler
 	bool updateResampleNote, updateVolFromText, updateVolToText, updateLPText;
 	bool updateHPText, updateNormFlag, update9xxPos;
 
@@ -255,7 +257,7 @@ typedef struct ui_t
 	int16_t lineCurX, lineCurY, editObject, sampleMarkingPos;
 	uint16_t *numPtr16, tmpDisp16, *dstOffset, dstPos, textLength, editTextPos;
 	uint16_t dstOffsetEnd, lastSampleOffset, diskOpPathTextOffset;
-	int32_t askTempData;
+	int32_t askTempData, *numPtr32, tmpDisp32;
 } ui_t;
 
 extern keyb_t keyb;
@@ -266,4 +268,4 @@ extern diskop_t diskop;
 extern cursor_t cursor;
 extern ui_t ui;
 
-extern module_t *song; // pt_main.c
+extern module_t *song; // pt2_main.c
