@@ -23,7 +23,6 @@
 #include "pt2_keyboard.h"
 #include "pt2_mouse.h"
 #include "pt2_audio.h"
-#include "pt2_palette.h"
 #include "pt2_helpers.h"
 #include "pt2_textout.h"
 #include "pt2_tables.h"
@@ -42,6 +41,8 @@
 #include "pt2_config.h"
 #include "pt2_bmp.h"
 #include "pt2_sampling.h"
+#include "pt2_chordmaker.h"
+#include "pt2_hpc.h"
 
 typedef struct sprite_t
 {
@@ -53,7 +54,6 @@ typedef struct sprite_t
 } sprite_t;
 
 static uint32_t vuMetersBg[4 * (10 * 48)];
-static uint64_t timeNext64, timeNext64Frac;
 
 sprite_t sprites[SPRITE_NUM]; // globalized
 
@@ -185,6 +185,12 @@ void statusSampleIsEmpty(void)
 	displayErrorMsg("SAMPLE IS EMPTY");
 }
 
+void changeStatusText(const char *text)
+{
+	fillRect(88, 127, 17*FONT_CHAR_W, FONT_CHAR_H, video.palette[PAL_GENBKG]);
+	textOut(88, 127, text, video.palette[PAL_GENTXT]);
+}
+
 void statusNotSampleZero(void)
 {
 	/* This rather confusing error message actually means that
@@ -194,67 +200,6 @@ void statusNotSampleZero(void)
 	displayErrorMsg("NOT SAMPLE 0 !");
 }
 
-void setupPerfFreq(void)
-{
-	uint64_t perfFreq64;
-	double dInt, dFrac;
-
-	perfFreq64 = SDL_GetPerformanceFrequency(); assert(perfFreq64 != 0);
-	editor.dPerfFreq = (double)perfFreq64;
-	editor.dPerfFreqMulMicro = 1000000.0 / editor.dPerfFreq;
-
-	// calculate vblank time for performance counters and split into int/frac
-	dFrac = modf(editor.dPerfFreq / VBLANK_HZ, &dInt);
-
-	// integer part
-	editor.vblankTimeLen = (int32_t)dInt;
-
-	// fractional part scaled to 0..2^32-1
-	dFrac *= UINT32_MAX;
-	dFrac += 0.5;
-	if (dFrac > UINT32_MAX)
-		dFrac = UINT32_MAX;
-	editor.vblankTimeLenFrac = (uint32_t)dFrac;
-}
-
-void setupWaitVBL(void)
-{
-	// set next frame time
-	timeNext64 = SDL_GetPerformanceCounter() + editor.vblankTimeLen;
-	timeNext64Frac = editor.vblankTimeLenFrac;
-}
-
-void waitVBL(void)
-{
-	// this routine almost never delays if we have 60Hz vsync, but it's still needed in some occasions
-
-	uint64_t time64 = SDL_GetPerformanceCounter();
-	if (time64 < timeNext64)
-	{
-		time64 = timeNext64 - time64;
-		if (time64 > UINT32_MAX)
-			time64 = UINT32_MAX;
-
-		const uint32_t diff32 = (uint32_t)time64;
-
-		// convert to microseconds and round to integer
-		const int32_t time32 = (int32_t)((diff32 * editor.dPerfFreqMulMicro) + 0.5);
-
-		// delay until we have reached the next frame
-		if (time32 > 0)
-			usleep(time32);
-	}
-
-	// update next tick time
-	timeNext64 += editor.vblankTimeLen;
-	timeNext64Frac += editor.vblankTimeLenFrac;
-	if (timeNext64Frac > 0xFFFFFFFF)
-	{
-		timeNext64Frac &= 0xFFFFFFFF;
-		timeNext64++;
-	}
-}
-
 void renderFrame(void)
 {
 	updateMOD2WAVDialog(); // must be first to avoid flickering issues
@@ -339,7 +284,7 @@ void showDownsampleAskDialog(void)
 	renderBigAskDialog();
 
 	textOutTight(133, 49, "THE SAMPLE'S FREQUENCY IS", video.palette[PAL_BACKGRD]);
-	textOutTight(178, 57, "ABOVE 22KHZ.", video.palette[PAL_BACKGRD]);
+	textOutTight(154, 57, "HIGH (ABOVE 22KHZ).", video.palette[PAL_BACKGRD]);
 	textOutTight(133, 65, "DO YOU WANT TO DOWNSAMPLE", video.palette[PAL_BACKGRD]);
 	textOutTight(156, 73, "BEFORE LOADING IT?", video.palette[PAL_BACKGRD]);
 }
@@ -488,26 +433,37 @@ void updateSongInfo1(void) // left side of screen, when Disk Op. is hidden
 	{
 		ui.updateCurrSampleLength = false;
 		if (!editor.isWAVRendering)
-			printFourHexBg(64, 69, *currSample->lengthDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+		{
+			if (config.maxSampleLength == 0xFFFE)
+				printFourHexBg(64, 69, *currSample->lengthDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+			else
+				printFiveHexBg(56, 69, *currSample->lengthDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+		}
 	}
 
 	if (ui.updateCurrSampleRepeat)
 	{
 		ui.updateCurrSampleRepeat = false;
-		printFourHexBg(64, 80, *currSample->loopStartDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+		if (config.maxSampleLength == 0xFFFE)
+			printFourHexBg(64, 80, *currSample->loopStartDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+		else
+			printFiveHexBg(56, 80, *currSample->loopStartDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 	}
 
 	if (ui.updateCurrSampleReplen)
 	{
 		ui.updateCurrSampleReplen = false;
-		printFourHexBg(64, 91, *currSample->loopLengthDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+		if (config.maxSampleLength == 0xFFFE)
+			printFourHexBg(64, 91, *currSample->loopLengthDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+		else
+			printFiveHexBg(56, 91, *currSample->loopLengthDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 	}
 }
 
 void updateSongInfo2(void) // two middle rows of screen, always present
 {
 	char tempChar;
-	int32_t secs, MI_TimeM, MI_TimeS, x, i;
+	int32_t x, i;
 	moduleSample_t *currSample;
 
 	if (ui.updateStatusText)
@@ -601,13 +557,13 @@ void updateSongInfo2(void) // two middle rows of screen, always present
 
 	// playback timer
 
-	secs = ((editor.musicTime >> 8) * 5) >> 9;
-	secs -= ((secs / 3600) * 3600);
+	const uint32_t ms1024 = editor.musicTime64 >> 32; // milliseconds (scaled from 1000 to 1024)
 
-	if (secs <= 5999) // below 99 minutes 59 seconds
+	uint32_t seconds = ms1024 >> 10;
+	if (seconds <= 5999) // below 100 minutes (99:59 is max for the UI)
 	{
-		MI_TimeM = secs / 60;
-		MI_TimeS = secs - (MI_TimeM * 60);
+		const uint32_t MI_TimeM = seconds / 60;
+		const uint32_t MI_TimeS = seconds - (MI_TimeM * 60);
 
 		// xx:xx
 		printTwoDecimalsBg(272, 102, MI_TimeM, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
@@ -709,9 +665,7 @@ void updateSampler(void)
 	{
 		if (!ui.samplerVolBoxShown && !ui.samplerFiltersBoxShown && s->length > 0)
 		{
-			tmpSampleOffset = (scr2SmpPos(mouse.x-3) + (1 << 7)) >> 8; // rounded
-			tmpSampleOffset = 0x900 + CLAMP(tmpSampleOffset, 0x00, 0xFF);
-
+			tmpSampleOffset = 0x900 + (scr2SmpPos(mouse.x-3) >> 8);
 			if (tmpSampleOffset != ui.lastSampleOffset)
 			{
 				ui.lastSampleOffset = (uint16_t)tmpSampleOffset;
@@ -724,7 +678,10 @@ void updateSampler(void)
 	if (ui.update9xxPos)
 	{
 		ui.update9xxPos = false;
-		printThreeHexBg(288, 247, ui.lastSampleOffset, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+		if (ui.lastSampleOffset <= 0x900 || ui.lastSampleOffset > 0x9FF)
+			textOutBg(288, 247, "---", video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+		else
+			printThreeHexBg(288, 247, ui.lastSampleOffset, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 	}
 
 	if (ui.updateResampleNote)
@@ -1236,6 +1193,15 @@ void renderEditOpScreen(void)
 
 	blit32(120, 44, 200, 55, srcPtr);
 
+	// fix graphics in 128K sample mode
+	if (config.maxSampleLength != 65534)
+	{
+		if (ui.editOpScreen == 2)
+			blit32(213, 55, 32, 11, fix128KPosBMP);
+		else if (ui.editOpScreen == 3)
+			blit32(120, 88, 48, 11, fix128KChordBMP);
+	}
+
 	renderEditOpMode();
 
 	// render text and content
@@ -1268,11 +1234,11 @@ void renderEditOpScreen(void)
 	{
 		textOut(128, 47, " SAMPLE CHORD EDITOR  ", video.palette[PAL_GENTXT]);
 
-		ui.updateLengthText = true;
-		ui.updateNote1Text = true;
-		ui.updateNote2Text = true;
-		ui.updateNote3Text = true;
-		ui.updateNote4Text = true;
+		ui.updateChordLengthText = true;
+		ui.updateChordNote1Text = true;
+		ui.updateChordNote2Text = true;
+		ui.updateChordNote3Text = true;
+		ui.updateChordNote4Text = true;
 	}
 }
 
@@ -1411,7 +1377,10 @@ void updateEditOp(void)
 		if (ui.updatePosText)
 		{
 			ui.updatePosText = false;
-			printFourHexBg(248, 58, *editor.samplePosDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+			if (config.maxSampleLength == 0xFFFE)
+				printFourHexBg(248, 58, *editor.samplePosDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+			else
+				printFiveHexBg(240, 58, *editor.samplePosDisp, video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 		}
 
 		if (ui.updateModText)
@@ -1435,13 +1404,17 @@ void updateEditOp(void)
 	}
 	else if (ui.editOpScreen == 3)
 	{
-		if (ui.updateLengthText)
+		if (ui.updateChordLengthText)
 		{
-			ui.updateLengthText = false;
+			ui.updateChordLengthText = false;
 
 			// clear background
-			textOutBg(168, 91, "    ", video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
-			charOut(198, 91,    ':', video.palette[PAL_GENBKG]);
+			if (config.maxSampleLength != 65534)
+				textOutBg(160, 91, "     ", video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+			else
+				textOutBg(168, 91, "    ", video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
+
+			charOut(198, 91, ':', video.palette[PAL_GENBKG]);
 
 			if (song->samples[editor.currSample].loopLength > 2 || song->samples[editor.currSample].loopStart >= 2)
 			{
@@ -1449,14 +1422,18 @@ void updateEditOp(void)
 			}
 			else
 			{
-				printFourHex(168, 91, *editor.chordLengthDisp, video.palette[PAL_GENTXT]); // chord max length
+				if (config.maxSampleLength == 0xFFFE)
+					printFourHex(168, 91, *editor.chordLengthDisp, video.palette[PAL_GENTXT]); // chord max length
+				else
+					printFiveHex(160, 91, *editor.chordLengthDisp, video.palette[PAL_GENTXT]); // chord max length
+
 				charOut(198, 91, (editor.chordLengthMin) ? '.' : ':', video.palette[PAL_GENTXT]); // min/max flag
 			}
 		}
 
-		if (ui.updateNote1Text)
+		if (ui.updateChordNote1Text)
 		{
-			ui.updateNote1Text = false;
+			ui.updateChordNote1Text = false;
 			if (editor.note1 > 35)
 				textOutBg(256, 58, "---", video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 			else
@@ -1464,9 +1441,9 @@ void updateEditOp(void)
 					video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 		}
 
-		if (ui.updateNote2Text)
+		if (ui.updateChordNote2Text)
 		{
-			ui.updateNote2Text = false;
+			ui.updateChordNote2Text = false;
 			if (editor.note2 > 35)
 				textOutBg(256, 69, "---", video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 			else
@@ -1474,9 +1451,9 @@ void updateEditOp(void)
 					video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 		}
 
-		if (ui.updateNote3Text)
+		if (ui.updateChordNote3Text)
 		{
-			ui.updateNote3Text = false;
+			ui.updateChordNote3Text = false;
 			if (editor.note3 > 35)
 				textOutBg(256, 80, "---", video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 			else
@@ -1484,9 +1461,9 @@ void updateEditOp(void)
 					video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 		}
 			
-		if (ui.updateNote4Text)
+		if (ui.updateChordNote4Text)
 		{
-			ui.updateNote4Text = false;
+			ui.updateChordNote4Text = false;
 			if (editor.note4 > 35)
 				textOutBg(256, 91, "---", video.palette[PAL_GENTXT], video.palette[PAL_GENBKG]);
 			else
@@ -1521,14 +1498,26 @@ void displayMainScreen(void)
 	if (ui.samplerScreenShown)
 	{
 		if (!ui.diskOpScreenShown)
+		{
 			blit32(0, 0, 320, 121, trackerFrameBMP);
+
+			if (config.maxSampleLength != 65534)
+				blit32(1, 65, 62, 34, fix128KTrackerBMP); // fix for 128kB support mode
+		}
 	}
 	else
 	{
 		if (!ui.diskOpScreenShown)
+		{
 			blit32(0, 0, 320, 255, trackerFrameBMP);
+
+			if (config.maxSampleLength != 65534)
+				blit32(1, 65, 62, 34, fix128KTrackerBMP); // fix for 128kB support mode
+		}
 		else
+		{
 			blit32(0, 121, 320, 134, &trackerFrameBMP[121 * SCREEN_W]);
+		}
 
 		ui.updateSongBPM = true;
 		ui.updateCurrPattText = true;
@@ -1623,13 +1612,7 @@ void handleAskNo(void)
 		break;
 
 		default:
-		{
 			restoreStatusAndMousePointer();
-			editor.errorMsgActive = true;
-			editor.errorMsgBlock = true;
-			editor.errorMsgCounter = 0;
-			setErrPointer();
-		}
 		break;
 	}
 
@@ -1762,7 +1745,7 @@ void handleAskYes(void)
 			s->loopLength = 2;
 
 			memset(s->text, 0, sizeof (s->text));
-			memset(&song->sampleData[(editor.currSample * MAX_SAMPLE_LEN)], 0, MAX_SAMPLE_LEN);
+			memset(&song->sampleData[(editor.currSample * config.maxSampleLength)], 0, config.maxSampleLength);
 
 			editor.samplePos = 0;
 			updateCurrSample();
@@ -2108,35 +2091,41 @@ void renderSprites(void)
 
 void flipFrame(void)
 {
+	const uint32_t windowFlags = SDL_GetWindowFlags(video.window);
+	bool minimized = (windowFlags & SDL_WINDOW_MINIMIZED) ? true : false;
+
 	renderSprites();
 	SDL_UpdateTexture(video.texture, NULL, video.frameBuffer, SCREEN_W * sizeof (int32_t));
-	SDL_RenderClear(video.renderer);
+
+	// SDL 2.0.14 bug on Windows (?): This function consumes ever-increasing memory if the program is minimized
+	if (!minimized)
+		SDL_RenderClear(video.renderer);
+
 	SDL_RenderCopy(video.renderer, video.texture, NULL, NULL);
 	SDL_RenderPresent(video.renderer);
 	eraseSprites();
 
 	if (!video.vsync60HzPresent)
 	{
-		waitVBL(); // we have no VSync, do crude thread sleeping to sync to ~60Hz
+		// we have no VSync, do crude thread sleeping to sync to ~60Hz
+		hpc_Wait(&video.vblankHpc);
 	}
 	else
 	{
-		uint32_t windowFlags = SDL_GetWindowFlags(video.window);
-
 		/* We have VSync, but it can unexpectedly get inactive in certain scenarios.
 		** We have to force thread sleeping (to ~60Hz) if so.
 		*/
 #ifdef __APPLE__
 		// macOS: VSync gets disabled if the window is 100% covered by another window. Let's add a (crude) fix:
-		if ((windowFlags & SDL_WINDOW_MINIMIZED) || !(windowFlags & SDL_WINDOW_INPUT_FOCUS))
-			waitVBL();
+		if (minimized || !(windowFlags & SDL_WINDOW_INPUT_FOCUS))
+			hpc_Wait(&video.vblankHpc);
 #elif __unix__
 		// *NIX: VSync gets disabled in fullscreen mode (at least on some distros/systems). Let's add a fix:
-		if ((windowFlags & SDL_WINDOW_MINIMIZED) || video.fullscreen)
-			waitVBL();
+		if (minimized || video.fullscreen)
+			hpc_Wait(&video.vblankHpc);
 #else
-		if (windowFlags & SDL_WINDOW_MINIMIZED)
-			waitVBL();
+		if (minimized)
+			hpc_Wait(&video.vblankHpc);
 #endif
 	}
 }
@@ -2144,31 +2133,30 @@ void flipFrame(void)
 void updateSpectrumAnalyzer(int8_t vol, int16_t period)
 {
 	const uint8_t maxHeight = SPECTRUM_BAR_HEIGHT + 1; // +1 because of audio latency - allows full height to be seen
-	int8_t scaledVol;
-	int32_t scaledNote;
+	uint16_t scaledVol;
+	uint32_t scaledNote;
 
 	if (ui.visualizerMode != VISUAL_SPECTRUM || vol <= 0)
 		return;
 
-	scaledVol = (vol * 24600L) >> 16; // scaledVol = (vol << 8) / 682
+	scaledVol = ((uint16_t)vol * 24576) >> 16; // scaledVol = vol / 2.66667 (0..64 -> 0..24)
 
 	period = CLAMP(period, 113, 856);
+	period -= 113;
 
-	scaledNote = 856 - period;
+	scaledNote = 743 - period;
 	scaledNote *= scaledNote;
-	scaledNote = ((int64_t)scaledNote * 171162) >> 32; // scaledNote /= 25093
-
-	// scaledNote now ranges 0..22, no need to clamp
+	scaledNote /= 25093; // scaledNote now ranges 0..22, no need to clamp
 
 	// increment main spectrum bar
-	editor.spectrumVolumes[scaledNote] += scaledVol;
+	editor.spectrumVolumes[scaledNote] += (uint8_t)scaledVol;
 	if (editor.spectrumVolumes[scaledNote] > maxHeight)
 		editor.spectrumVolumes[scaledNote] = maxHeight;
 
 	// increment left side of spectrum bar with half volume
 	if (scaledNote > 0)
 	{
-		editor.spectrumVolumes[scaledNote-1] += scaledVol >> 1;
+		editor.spectrumVolumes[scaledNote-1] += (uint8_t)(scaledVol >> 1);
 		if (editor.spectrumVolumes[scaledNote-1] > maxHeight)
 			editor.spectrumVolumes[scaledNote-1] = maxHeight;
 	}
@@ -2176,7 +2164,7 @@ void updateSpectrumAnalyzer(int8_t vol, int16_t period)
 	// increment right side of spectrum bar with half volume
 	if (scaledNote < SPECTRUM_BAR_NUM-1)
 	{
-		editor.spectrumVolumes[scaledNote+1] += scaledVol >> 1;
+		editor.spectrumVolumes[scaledNote+1] += (uint8_t)(scaledVol >> 1);
 		if (editor.spectrumVolumes[scaledNote+1] > maxHeight)
 			editor.spectrumVolumes[scaledNote+1] = maxHeight;
 	}
@@ -2189,7 +2177,7 @@ void sinkVisualizerBars(void)
 	// sink stuff @ 49.92Hz (Amiga PAL) rate
 
 	static uint64_t counter50Hz;
-	const uint64_t counter50HzDelta = (uint64_t)(((UINT32_MAX + 1.0) * (AMIGA_PAL_VBLANK_HZ / (double)VBLANK_HZ)) + 0.5);
+	const uint64_t counter50HzDelta = (uint64_t)(((UINT32_MAX+1.0) * (AMIGA_PAL_VBLANK_HZ / (double)VBLANK_HZ)) + 0.5);
 
 	counter50Hz += counter50HzDelta; // 32.32 fixed-point counter
 	if (counter50Hz > 0xFFFFFFFF)
@@ -2214,21 +2202,24 @@ void sinkVisualizerBars(void)
 
 void updateRenderSizeVars(void)
 {
-	int32_t di;
-#ifdef __APPLE__
-	int32_t actualScreenW, actualScreenH;
-	double dXUpscale, dYUpscale;
-#endif
 	float fXScale, fYScale;
-	SDL_DisplayMode dm;
 
-	di = SDL_GetWindowDisplayIndex(video.window);
-	if (di < 0)
-		di = 0; /* return display index 0 (default) on error */
+	if (video.useDesktopMouseCoords)
+	{
+		SDL_DisplayMode dm;
 
-	SDL_GetDesktopDisplayMode(di, &dm);
-	video.displayW = dm.w;
-	video.displayH = dm.h;
+		int32_t di = SDL_GetWindowDisplayIndex(video.window);
+		if (di < 0)
+			di = 0; // return display index 0 (default) on error
+
+		SDL_GetDesktopDisplayMode(di, &dm);
+		video.displayW = dm.w;
+		video.displayH = dm.h;
+	}
+	else
+	{
+		SDL_GetWindowSize(video.window, &video.displayW, &video.displayH);
+	}
 
 	if (video.fullscreen)
 	{
@@ -2246,20 +2237,25 @@ void updateRenderSizeVars(void)
 			video.renderW = (int32_t)(SCREEN_W * fXScale);
 			video.renderH = (int32_t)(SCREEN_H * fYScale);
 
-#ifdef __APPLE__
-			// retina high-DPI hackery (SDL2 is bad at reporting actual rendering sizes on macOS w/ high-DPI)
+			// high-DPI hackery:
+			//  On high-DPI systems, the display w/h are given in logical pixels,
+			//  but the renderer size is given in physical pixels. Since our internal
+			//  render{X,Y,W,H} variables need to be in logical coordinates, as that's
+			//  what mouse input uses, scale them by the screen's DPI scale factor,
+			//  which is the physical (renderer) size / the logical (window) size.
+			//  On non high-DPI systems, this is effectively a no-op.
+			int32_t actualScreenW, actualScreenH;
 			SDL_GL_GetDrawableSize(video.window, &actualScreenW, &actualScreenH);
-			SDL_GetDesktopDisplayMode(0, &dm);
 
-			dXUpscale = ((double)actualScreenW / video.displayW);
-			dYUpscale = ((double)actualScreenH / video.displayH);
+			const double dXUpscale = (const double)actualScreenW / video.displayW;
+			const double dYUpscale = (const double)actualScreenH / video.displayH;
 
 			// downscale back to correct sizes
 			if (dXUpscale != 0.0) video.renderW = (int32_t)(video.renderW / dXUpscale);
 			if (dYUpscale != 0.0) video.renderH = (int32_t)(video.renderH / dYUpscale);
-#endif
-			video.renderX = (video.displayW - video.renderW) >> 1;
-			video.renderY = (video.displayH - video.renderH) >> 1;
+
+			video.renderX = (video.displayW - video.renderW) / 2;
+			video.renderY = (video.displayH - video.renderH) / 2;
 		}
 	}
 	else
@@ -2271,21 +2267,29 @@ void updateRenderSizeVars(void)
 	}
 
 	// for mouse cursor creation
-	video.xScale = (int32_t)((video.renderW * (1.0 / SCREEN_W)) + 0.5);
-	video.yScale = (int32_t)((video.renderH * (1.0 / SCREEN_H)) + 0.5);
+	video.xScale = (int32_t)round(video.renderW / (double)SCREEN_W);
+	video.yScale = (int32_t)round(video.renderH / (double)SCREEN_H);
+
 	createMouseCursors();
 }
 
-void toggleFullScreen(void)
+void toggleFullscreen(void)
 {
-	SDL_DisplayMode dm;
-
 	video.fullscreen ^= 1;
 	if (video.fullscreen)
 	{
+		SDL_SetWindowFullscreen(video.window, SDL_WINDOW_FULLSCREEN_DESKTOP);
+
 		if (config.fullScreenStretch)
 		{
-			SDL_GetDesktopDisplayMode(0, &dm);
+			SDL_DisplayMode dm;
+
+			int32_t di = SDL_GetWindowDisplayIndex(video.window);
+			if (di < 0)
+				di = 0; // return display index 0 (default) on error
+
+			SDL_GetDesktopDisplayMode(di, &dm);
+
 			SDL_RenderSetLogicalSize(video.renderer, dm.w, dm.h);
 		}
 		else
@@ -2294,16 +2298,24 @@ void toggleFullScreen(void)
 		}
 
 		SDL_SetWindowSize(video.window, SCREEN_W, SCREEN_H);
-		SDL_SetWindowFullscreen(video.window, SDL_WINDOW_FULLSCREEN_DESKTOP);
+
+#ifndef __unix__ // can be severely buggy on Linux... (at least when used like this)
 		SDL_SetWindowGrab(video.window, SDL_TRUE);
+#endif
 	}
 	else
 	{
 		SDL_SetWindowFullscreen(video.window, 0);
 		SDL_RenderSetLogicalSize(video.renderer, SCREEN_W, SCREEN_H);
 		SDL_SetWindowSize(video.window, SCREEN_W * config.videoScaleFactor, SCREEN_H * config.videoScaleFactor);
+
+#ifdef __unix__ // can be required on Linux... (or else the window keeps moving down every time you leave fullscreen)
 		SDL_SetWindowPosition(video.window, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED);
+#endif
+
+#ifndef __unix__
 		SDL_SetWindowGrab(video.window, SDL_FALSE);
+#endif
 	}
 
 	updateRenderSizeVars();
@@ -2311,13 +2323,13 @@ void toggleFullScreen(void)
 
 	if (video.fullscreen)
 	{
-		mouse.setPosX = video.displayW >> 1;
-		mouse.setPosY = video.displayH >> 1;
+		mouse.setPosX = video.displayW / 2;
+		mouse.setPosY = video.displayH / 2;
 	}
 	else
 	{
-		mouse.setPosX = video.renderW >> 1;
-		mouse.setPosY = video.renderH >> 1;
+		mouse.setPosX = video.renderW / 2;
+		mouse.setPosY = video.renderH / 2;
 	}
 
 	mouse.setPosFlag = true;
@@ -2325,14 +2337,10 @@ void toggleFullScreen(void)
 
 bool setupVideo(void)
 {
-	int32_t screenW, screenH;
-	uint32_t rendererFlags;
-	SDL_DisplayMode dm;
+	int32_t screenW = SCREEN_W * config.videoScaleFactor;
+	int32_t screenH = SCREEN_H * config.videoScaleFactor;
 
-	screenW = SCREEN_W * config.videoScaleFactor;
-	screenH = SCREEN_H * config.videoScaleFactor;
-
-	rendererFlags = 0;
+	uint32_t rendererFlags = 0;
 
 #ifdef _WIN32
 #if SDL_PATCHLEVEL >= 4
@@ -2340,14 +2348,16 @@ bool setupVideo(void)
 #endif
 #endif
 
-#if SDL_PATCHLEVEL >= 5
-	SDL_SetHint(SDL_HINT_MOUSE_FOCUS_CLICKTHROUGH, "1");
-#endif
-
 	video.vsync60HzPresent = false;
 	if (!config.vsyncOff)
 	{
-		SDL_GetDesktopDisplayMode(0, &dm);
+		SDL_DisplayMode dm;
+
+		int32_t di = SDL_GetWindowDisplayIndex(video.window);
+		if (di < 0)
+			di = 0; // return display index 0 (default) on error
+
+		SDL_GetDesktopDisplayMode(di, &dm);
 		if (dm.refresh_rate >= 59 && dm.refresh_rate <= 61)
 		{
 			video.vsync60HzPresent = true;
@@ -2420,6 +2430,12 @@ bool setupVideo(void)
 	// we want an aligned pointer
 	video.frameBuffer = (uint32_t *)ALIGN_PTR(video.frameBufferUnaligned, 256);
 
+	// Workaround: SDL_GetGlobalMouseState() doesn't work with KMSDRM/Wayland
+	video.useDesktopMouseCoords = true;
+	const char *videoDriver = SDL_GetCurrentVideoDriver();
+	if (videoDriver != NULL && (strcmp("KMSDRM", videoDriver) == 0 || strcmp("wayland", videoDriver) == 0))
+		video.useDesktopMouseCoords = false;
+
 	updateRenderSizeVars();
 	updateMouseScaling();
 
@@ -2428,11 +2444,6 @@ bool setupVideo(void)
 	else
 		SDL_ShowCursor(SDL_FALSE);
 
-	// Workaround: SDL_GetGlobalMouseState() doesn't work with KMSDRM
-	video.useDesktopMouseCoords = true;
-	const char *videoDriver = SDL_GetCurrentVideoDriver();
-	if (videoDriver != NULL && strcmp("KMSDRM", videoDriver) == 0)
-		video.useDesktopMouseCoords = false;
-
+	SDL_SetRenderDrawColor(video.renderer, 0, 0, 0, SDL_ALPHA_OPAQUE);
 	return true;
 }
